/* cmdline.c
   Program for general command line interface
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "common.h"

/* Extra includes used by this file only */
#include <ctype.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/stat.h>

/* Buffer for output strings (linebreaking etc) */
static char Buff[BUFFSIZE];
static bool OldLineBreakMode;
static bool OldPreserveWhitespace;

static char *cLangCodes[] = {"SV","EN","ES"};
static char *cLanguages[] = {"Swedish","English","Spanish"};
static char *cContents[]  = {"Innehll","Contents","Contents"};
static char *cImage[]     = {"Bild", "Image", "Image"};
static char *cFootnotes[] = {"Fotnoter", "Footnotes", "Footnotes"};

/* OutWrap is called by the functions below to wrap lines in output */
static void
OutWrap(void)
{
  int i,j,k;
  char tmp;
  char tmpBuffer[160];

  /* Dont process an empty buffer */
  if (!strlen(LineBuffer))
    return;

  /* Search backwards til you reach a ' ' */
   i = j = strlen(LineBuffer)-1;
  while((i!=0) && (LineBuffer[i]!=' '))
    i--;
  /* If there are no possible delimiters in the buffer, just print it */
  if (i==0)
    fprintf(OutFile, LineBuffer);
  /* Otherwise print up to the desisired point */
  else {
    tmp = LineBuffer[i];
    LineBuffer[i] = '\0';
    fprintf(OutFile, LineBuffer);
    /* The space should actually be written */
    fputc(tmp, OutFile);
    /* And then wrap the line with a linefeed */
    fputc('\n', OutFile);
    /* Then copy the remainder to the beginning of the buffer          */
    /* But only if the space or newline was not the last character     */
    /* of the string. Possibly this could be done in one operation ... */
    if (i < j)
      strcpy(tmpBuffer, LineBuffer+i+1);
    /* If we are to indent, then indent, else just zap the string */
    if (IndentChars) {
      for (k=0;k<IndentChars;k++)
	LineBuffer[k] = ' ';
      LineBuffer[k] = '\0';
    } else {
      LineBuffer[0] = '\0';
    }
    if (i < j)
      strcat(LineBuffer, tmpBuffer);
  }
}

/************************************************************************
 * OutString and OutChar are exported "public" functions which are
 * called by the different converters for buffering output (text) data.
 ************************************************************************/

void
SetStackMode()
{
  /*
    StackMode + LineBreakMode may not be set simultaneously,
    linebreaking of the stack is made after all has been written
    to the main file.
  */
  StackMode = true;
  OldLineBreakMode = LineBreakMode;
  OldPreserveWhitespace = PreserveWhitespace;
  LineBreakMode = false;
}

void
UnSetStackMode()
{
  StackMode = false;
  LineBreakMode = OldLineBreakMode;
  PreserveWhitespace = OldPreserveWhitespace;
}

/*
 * Puts language code to lower char and
 * returns it. Also corrects the error of
 * writing "se" instead of "sv"
 */

char *
cFixLangCode(const char *inStr)
{
  char *newStr;
  register int i;
  
  newStr = (char *) malloc(strlen(inStr)+1);
  strcpy(newStr, inStr);
  for (i=0; i<strlen(newStr); i++)
	 newStr[i] = (char) toupper(newStr[i]);
  if (!strcmp(newStr,"SE"))
    strcpy(newStr, "SV");
  return (newStr);
}

/*
 * Looks up a language from a language code and
 * returns a pointer to a long language name
 * "sv" -> "Swedish" etc.
 */
char *
cGetLLanguage(const char *cLangCode)
{
  register int i;

  if (!cLangCode)
    return(cLanguages[0]);
  for (i=0; i < 3; i++) {
    if (!strcmp(cLangCodes[i],cLangCode))
      return(cLanguages[i]);
  }
  return (cLanguages[0]);
}

char *
cGetLContents(const char *cLangCode)
{
  register int i;

  if (!cLangCode)
    return(cContents[0]);
  for (i=0; i < 3; i++) {
    if (!strcmp(cLangCodes[i],cLangCode))
      return(cContents[i]);
  }
  return (cContents[0]);
}

char *
cGetLImage(const char *cLangCode)
{
  register int i;

  if (!cLangCode)
    return(cImage[0]);
  for (i=0; i < 3; i++) {
    if (!strcmp(cLangCodes[i],cLangCode))
      return(cImage[i]);
  }
  return (cImage[0]);
}

char *
cGetLFootnotes(const char *cLangCode)
{
  register int i;

  if (!cLangCode)
    return(cFootnotes[0]);
  for (i=0; i < 3; i++) {
    if (!strcmp(cLangCodes[i],cLangCode))
      return(cFootnotes[i]);
  }
  return (cFootnotes[0]);
}

void
OutString(char *fmt, ...)
{
  char tmpBuffer[160];
  va_list ap;
  register int i;

  va_start(ap, fmt);
  if (!LineBreakMode) {
    /* This routine just passes the string through */
    if (StackMode)
      vfprintf(FootnoteStackFile, fmt, ap);
    else
      vfprintf(OutFile, fmt, ap);
  } else {
    /* This one breaks the lines */
     vsprintf(tmpBuffer, fmt, ap);
     for (i=0; i<strlen(tmpBuffer); i++)
       OutChar(tmpBuffer[i]);
  }
  va_end(ap);
}

void
OutChar(char c)
{
  int BuffLen;
  int i;

  if (!LineBreakMode) {
    /* This routine just passes the char through */
    if (StackMode)
      fputc(c, FootnoteStackFile);
    else
      fputc(c, OutFile);
  } else {
    /* This one breaks the lines */
    BuffLen=strlen(LineBuffer);
    LineBuffer[BuffLen] = c;
    LineBuffer[BuffLen+1] = '\0';
    /* String functions may treat %-signs harshly */
    if (c == '%') {
      LineBuffer[BuffLen+1] = '%';
      LineBuffer[BuffLen+2] = '\0';
    }
    if (c == '\n') {
      fprintf(OutFile, LineBuffer);
      LineBuffer[0] = '\0';
      if (IndentChars) {
	for (i=0;i<IndentChars;i++)
	  LineBuffer[i] = ' ';
	LineBuffer[i] = '\0';
      }
    } else if (strlen(LineBuffer) > 78)
      OutWrap();
  }
}

/************************************************************************
 * Recursively search files and directories below the directory i
 * "base_path" for the file "name". Return a open file pointer to the
 * file in question or NULL if the search failed.
 ************************************************************************/
static FILE *
FileScan(char * base_path, char * name)
{
  DIR * DirPointer;
  struct dirent * Entry;
  struct stat EntryStatus;
  char cTempPath[512];
  int bIsDir;
  int bIsFile;
  FILE * ReturnPointer=NULL;
  FILE * TempPointer=NULL;

  DirPointer = opendir(base_path);
  if (! DirPointer){
    fprintf(stderr, "Unable to open directory %s.\n", base_path);
    return NULL;
  }
  while ((Entry = readdir(DirPointer)) != NULL)
    {
      strncpy(cTempPath, base_path, 512);
      strcat(cTempPath, "/");
      strcat(cTempPath, Entry->d_name);
      if (stat(cTempPath, &EntryStatus)) {
	fprintf(stderr, "Failed to read file status for %s.\n", cTempPath);
	return 0;
      }
      bIsDir = S_ISDIR(EntryStatus.st_mode);
      bIsFile = S_ISREG(EntryStatus.st_mode);
      if (bIsDir && strcmp(Entry->d_name, ".") && strcmp(Entry->d_name,"..") ) {
	TempPointer = FileScan(cTempPath, name);
	if ((ReturnPointer==NULL) && (TempPointer!=NULL)) {
	  ReturnPointer = TempPointer;
	  closedir(DirPointer);
	  return ReturnPointer;
	}
      }
      if (bIsFile) {
	if (strcmp(Entry->d_name, name) == 0) {
	  /* Stuff to do when found */
	  if (ReturnPointer==NULL) {
	    ReturnPointer=fopen(cTempPath,"r");
	    closedir(DirPointer);
	    return ReturnPointer;
	  }
	}
      }
    }
  closedir(DirPointer);
  return ReturnPointer;
}

/************************************************************************
 * Main function call to XML translators
 ************************************************************************/
int
main(const int argc, const char **argv) {
  /* Variables */
  XML_Parser parser;
  bool done;
  int len;
  int arg;
  /* Signifies if the program should write to stdout instead of a file */
  bool Stdout = false;
  bool DirSearch = false;
  char cDir[STR_MAXLEN] = "";
  char cInFile[STR_MAXLEN] = "";
  char cOutFile[STR_MAXLEN] = "";
  static char cProgramName[]= PACKAGE "-" VERSION;
  FILE* InFile;

  if (argc<2) {
    printf("Usage: %s [-options] <file.xml>\n", argv[0]);
    printf("Type %s -help for all options.\n", argv[0]);
    exit(0);
  }

  /* Set some global variables b4 we start */
  /* Do not print ANYTHING until we raise this flag */
  CharThroughPut = false;
  /* We want to wrap text by default */
  LineBreakMode = true;
  /* Clear the common line buffer */
  LineBuffer[0] = '\0';
  /* No indentation by default */
  IndentChars = 0;
  /* Turn off and clear stacking mode (stacking of footnotes) */
  StackMode = false;
  FootnoteNo = 0;
  /* As default, do not preserve any whitespace */
  PreserveWhitespace = false;
  /* Default to no chapter to extract */
  ExtractChapter[0] = '\0';
  /* Do not extract a TOC */
  ExtractToc = false;
  /* Default is to go over file just once */
  XMLIterations = 1;
  /* Set program name, and recoding engines NULL */
  program_name = cProgramName;
  recode_engine_outer = NULL;
  recode_engine_request_1 = NULL;
  recode_engine_request_2 = NULL;

  for (arg=1; arg<argc; arg++) {
    if (*argv[arg] == '-') {
      /* -? or -help: show the command line help page */
      if (!strcmp(argv[arg],"-?") || !strcmp(argv[arg],"-help") || !strcmp(argv[arg],"--help")) {
	printf("Usage: %s [-options] <file[.xml]>\n\n", argv[0]);
	printf("Options: -             = Write to stdout\n");
	printf("         -R <dir>      = Recursively search this <dir> for <file[.xml]>\n");
	printf("         -c <chap>     = Extract this chapter only (for books)\n");
	printf("         -toc          = Create a Table of Contents (need named headings)\n");
	/* Print any tool-specific help text */
	PrintToolHelp();
	exit(0);
	/* non-interactive */
      }
      if (!strcmp(argv[arg],"-"))
	Stdout = true;
      if (!strcmp(argv[arg],"-R")) {
	DirSearch = true;
	if(*argv[arg+1] == '-') {
	  fprintf(stderr, "You must give a <directory> parameter to the -R switch!\n");
	  exit(-1);
	}
	strcpy(cDir, argv[arg+1]);
	if(!Stdout)
	  fprintf(stderr, "Using XML source tree at: %s\n", cDir);
      }
      if (!strcmp(argv[arg],"-c")) {
	if(*argv[arg+1] == '-') {
	  fprintf(stderr, "You must give a <chapter> parameter to the -c switch!\n");
	  exit(-1);
	}
	strncpy(ExtractChapter, argv[arg+1], 127);
	ExtractChapter[127] = '\0';
	if (!Stdout)
	  fprintf(stderr, "Extracting chapter: %s\n", ExtractChapter);
      }
      /* Checks for tool-specific options (implemented in all subtools) */
      ScanToolOptions(argv, arg);
    }
  }

  /* See that the source is OK */
  strcpy(cInFile, argv[argc-1]);
  if (strcmp(cInFile+(strlen(cInFile)-4), ".xml"))
    strcpy(cInFile+strlen(cInFile), ".xml");
  
  if(!Stdout) {
    strcpy(cOutFile, cInFile);
    /* 
       This is handy. Get extension from program name. Ad hoc?
       If xmltohtml is run, add .html and if it is not given
       a certain chapter, then run ToC generation first by
       force, then normal parsing.
    */
    if (!strcmp(argv[0]+strlen(argv[0])-4, "html")) {
      strcpy(cOutFile+strlen(cOutFile)-3,"html");
    } else {
      strcpy(cOutFile+strlen(cOutFile)-3,argv[0]+strlen(argv[0])-3);
    }
    fprintf(stderr, "Input file: %s\n", cInFile);
    fprintf(stderr, "Output file: %s\n", cOutFile);
  }

  /* Scan for the file or just assume it is here and open it */
  if (DirSearch)
    InFile = FileScan(cDir, cInFile);
  else
    InFile = fopen(cInFile, "r");
  if (InFile == NULL) {
    if (DirSearch)
      fprintf(stderr, "Could not find <%s> in <%s> !\n", cInFile, cDir);
    else
      fprintf(stderr, "Could not open input file <%s!>\n", cInFile);
    exit(-1);
  }

  if(!Stdout) {
    OutFile = fopen(cOutFile, "w");
    if (OutFile == NULL) {
      fprintf(stderr, "Could not create output file <%s>!\n", cOutFile);
      exit(-1);
    }
  } else {
    OutFile = stdout;
  }

  while (XMLIterations--) {

    parser = XML_ParserCreate(NULL);
    
    if (! parser) {
      fprintf(stderr, "Could not create an XML parser instance\n");
      exit(-1);
    }

    /* Create an XML parser */
    XML_SetElementHandler(parser, startElementHandler, endElementHandler);
    XML_SetCharacterDataHandler(parser, textHandler);
    XML_SetDefaultHandler(parser, defaultHandler);

    if (fseek(InFile, 0, SEEK_SET)) {
      fprintf(stderr, "Could not read from input file <%s>!\n", cInFile);
      exit(-1);
    }
    done = false;
    while (!done) {
      len = fread(Buff, 1, BUFFSIZE, InFile);
      if (ferror(InFile)) {
	fprintf(stderr, "Could not read from input file <%s>!\n", cInFile);
	exit(-1);
      }
      done = feof(InFile);
      
      if (! XML_Parse(parser, Buff, len, done)) {
	fprintf(stderr, "Parse error at line %d:\n%s\n",
		XML_GetCurrentLineNumber(parser),
		XML_ErrorString(XML_GetErrorCode(parser)));
	XML_ParserFree(parser);
	exit(-1);
      }
    }
    XML_ParserFree(parser);
  }

  /* Free recode engine */
  if (recode_engine_request_1 != NULL)
    recode_delete_request(recode_engine_request_1);
  if (recode_engine_request_2 != NULL)
    recode_delete_request(recode_engine_request_2);
  if (recode_engine_outer != NULL)
    recode_delete_outer(recode_engine_outer);

  fclose(InFile);
  OutChar('\n');
  if (!Stdout)
    fclose(OutFile);
  exit(0);
}
