/* lcd2html.c
   Program for parsing and converting article.dtd or
   book.dtd compliant XML files to HTML 4.0 (Transitional)
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the LCDtools package.

Yelah LCDtools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "common.h"

/* Language identifier "sv"=Swedish, "en"=English etc... */
char *cLangCode;
/* Parse a date */
bool ParseDate = false;
/* This one is used to remove extraneous space in character printing */
static unsigned char LastChar = ' ';
/* For handling of headerlessness */
bool Header = true;
/* Disable graphics by default */
bool EnableGraphics = false;
/* Produce XHTML, yes/no? */
bool ProduceXHTML = false;
/* Virtual path to use on graphics (URI:s) */
char GraphicsVirtualPath[128] = "\0";
/* A flag and chapter counter for books */
bool bIsBook = false;
int ChapCount = 0;
/* Counts headings for generating ToC */
bool ExtractedToc = false;
bool IgnoreFootnotes = false;
bool oldCharThroughPut = false;
bool HasNameAttrib = false;
int h1Count = 0;
int h2Count = 0;
int h3Count = 0;

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  register int i, j, k;
  char cTmp[128];

  if (!strcmp(el,"article")) {
    bIsBook = false;
    if (Header) {
      if (ProduceXHTML) {
	OutString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	OutString("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n");
	OutString(" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n");
	OutString("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"sv\" lang=\"sv\">\n<head>\n");
      } else
	OutString("<!DOCTYPE HTML PUBLIC \"-//W3C/DTD HTML 4.01 Transitional//EN\">\n<html>\n<head>\n");
      CharThroughPut = true;
    }
  }
  if (!strcmp(el,"book")) {
    bIsBook = true;
    if (Header) {
      if (ProduceXHTML) {
	OutString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	OutString("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n");
	OutString(" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n");
	OutString("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"sv\" lang=\"sv\">\n<head>\n");
      } else
	OutString("<!DOCTYPE HTML PUBLIC \"-//W3C/DTD HTML 4.01 Transitional//EN\">\n<html>\n<head>\n");
      CharThroughPut = true;
    }
  }
  else if (!strcmp(el,"metadata") && Header) {
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "cathegory")) {
	/* Totally ignore cathegory */
      }
      if (!strcmp(attr[i], "language")) {
	cLangCode = cFixLangCode(attr[i+1]);
	if (ProduceXHTML) {
	  OutString("<meta http-equiv=\"Content-Type\" content=\"application/xhtml+xml\" />\n");
	} else {
	  OutString("<meta http-equiv=\"Content-Type\" content=\"text/html\" charset=\"ISO-8859-1\" lang=\"%s\">\n", cLangCode);
	  OutString("<meta http-equiv=\"Content-Language\" content=\"%s\">\n", cLangCode);
	  OutString("<meta name=\"language\" content=\"%s\">\n", cGetLLanguage(cLangCode));
	}
      }
    }
    OutString("<meta name=\"resource-type\" content=\"document\"");
    OutString(ProduceXHTML ? " />\n" : ">\n");
    OutString("<meta name=\"distribution\" content=\"global\"");
    OutString(ProduceXHTML ? " />\n" : ">\n");
  }
  else if (!strcmp(el,"description") && Header) {
    PreserveWhitespace = true;
    OutString("<title>");
  }
  else if (!strcmp(el,"author") && Header) {
    PreserveWhitespace = true;
    OutString("<meta name=\"author\" content=\"");
  }
  else if (!strcmp(el,"keywords") && Header) {
    PreserveWhitespace = true;
    OutString("<meta name=\"keywords\" content=\"");
  }
  else if (!strcmp(el,"published") && Header) {
    ParseDate = true;
    OutString("<meta name=\"copyright\" content=\"");
  }
  /* Body handling for articles */
  else if (!strcmp(el,"body")) {
    /* I just preserve all whitespace in the body... Its HTML anyway! */
    PreserveWhitespace = true;
    CharThroughPut = true;
    /* If in headerless mode, char throughput begins HERE, else just write the start body tag */
    if (Header)
      if (ProduceXHTML) {
	OutString("<body style=\"color: black; background: white\">\n");
      }else {
	OutString("<body bgcolor=\"#ffffff\">\n");
      }
    LastChar = ' ';
  }
  /* Chapter handling for books */
  else if (!strcmp(el,"chapter")) {
    /* If we're extracting with no chapter given, extract a TOC then the full book */

    if (!ChapCount && !strlen(ExtractChapter) && ExtractToc && !ExtractedToc) {
      /* Reiterate file for the book text */
      XMLIterations++;
    }
    if (ExtractToc) {
      if (Header && !ChapCount) {
	if (ProduceXHTML) {
	  OutString("<body style=\"color: black; background: white\">\n");
	}else {
	  OutString("<body bgcolor=\"#ffffff\">\n");
	}
      }
      LastChar = ' ';
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "label")) {
	  /* label is in attr[i+1], unused */
	}
      }
    }
    if (strlen(ExtractChapter)) {
      /* Compare and extract it on match */
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "label")) {
	  if (!strcmp(ExtractChapter, attr[i+1])) {
	    PreserveWhitespace = true;
	    CharThroughPut = true;
	    if (Header) {
	      if (ProduceXHTML) {
		OutString("<body style=\"color: black; background: white\">\n");
	      }else {
		OutString("<body bgcolor=\"#ffffff\">\n");
	      }
	    }
	    OutString("<a name=\"%s\"></a>\n", attr[i+1]);
	  }
	}
      }
    }
    if (!ExtractToc && !strlen(ExtractChapter)) {
      /* If no extraction chapter is given, extract all chapters */
      PreserveWhitespace = true;
      CharThroughPut = true;
      if (Header && !ChapCount) {
	if (ProduceXHTML) {
	  OutString("<body style=\"color: black; background: white\">\n");
	}else {
	  OutString("<body bgcolor=\"#ffffff\">\n");
	}
      }
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "label")) {
	  OutString("<a name=\"%s\"></a>\n", attr[i+1]);
	}
      }
      /* Separate chapters, except for the first one */
      if (ChapCount)
	OutString(ProduceXHTML ? "<hr />\n" : "<ht>\n");
    }
    ChapCount++;
  }
  /* Elements to be copied with a justify attribute */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")
	   ) {

    /* Routine to extract ToC from books */
    if (bIsBook && ExtractToc) {
      /* Find a name attribute */
      HasNameAttrib = false;
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "name")) {
	  HasNameAttrib = true;
	  strncpy(cTmp, attr[i+1],127);
	  cTmp[127] = '\0';
	}
      }
      CharThroughPut = true; /* To be reset by end tag if ExtractToc is true */
      PreserveWhitespace = true;
      if (!strcmp(el,"h1")) {
	if (h3Count) {
	  OutString("</ul>\n");
	  h3Count = 0;
	}
	if (h2Count) {
	  OutString("</ul>\n");
	  h2Count = 0;
	}
	if (!h1Count) { /* Begin ToC on first occurence */
	  OutString("<h1>%s</h1>\n", cGetLContents(cLangCode));
	  OutString("<ul>\n");
	}
	h1Count++;
      }
      if (!strcmp(el,"h2")) {
	if (h3Count) {
	  OutString("</ul>\n");
	  h3Count = 0;
	}
	if (!h2Count)
	  OutString("<ul>\n");
	h2Count++;
      }
      if (!strcmp(el,"h3")) {
	if (!h3Count)
	  OutString("<ul>\n");
	h3Count++;
      }
      OutString("<li>");
      if (HasNameAttrib)
	OutString("<a href=\"#%s\">", cTmp);
    }

    /* If we were NOT producing a ToC, continue here */
    else if (CharThroughPut) {
      OutString("<%s", el);
      for (i = 0; attr[i]; i += 2) {
	HasNameAttrib = false;
	if (!strcmp(attr[i], "align"))
	  if (ProduceXHTML) {
	    OutString(" style=\"text-align: %s\"", attr[i+1]);
	  } else {
	    OutString(" align=\"%s\"", attr[i+1]);
	  }
	if (!strcmp(attr[i], "name")) {
	  HasNameAttrib = true;
	  strncpy(cTmp,attr[i+1],127);
	  cTmp[127] = '\0';
	}
      }
      OutString(">");
      if (HasNameAttrib)
	OutString("<a name=\"%s\">", cTmp);
    }
  }

  else if (CharThroughPut && !strcmp(el,"p")) {
    OutString("<p");
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align"))
	if (ProduceXHTML) {
	  OutString(" style=\"text-align: %s\"", attr[i+1]);
	} else {
	  OutString(" align=\"%s\"", attr[i+1]);
	}
    }
    OutString(">");
  }
  /* Elements to be plain copied WITH ENTER after them */
  else if (
	   CharThroughPut &&
	   /* "If I am not busy extracting a ToC" */
	   !(bIsBook && (ExtractToc && !ExtractedToc)) &&
	   (!strcmp(el,"br") ||
	   !strcmp(el,"hr"))
	   )
    if (ProduceXHTML)
      OutString("<%s />", el);
    else
      OutString("<%s>", el);
  /* To be XHTML compliant */
  else if (
	   CharThroughPut && (
	   !strcmp(el,"blockquote") ||
	   !strcmp(el,"table") ||
	   !strcmp(el,"tr") ||
	   !strcmp(el,"ol") ||
	   !strcmp(el,"ul") ||
	   !strcmp(el,"dl"))
	   )
    OutString("<%s>\n", el);
     
  /* Not using valign="top" on table cells results in severe trauma */
  else if (CharThroughPut && !strcmp(el,"td"))
    OutString("<td valign=\"top\">\n");

  /* Elements to be plain copied with NO ENTER after them */
  else if (
	   CharThroughPut && (
	   !strcmp(el,"b") ||
	   !strcmp(el,"i") ||
	   !strcmp(el,"li") ||
	   !strcmp(el,"dt") ||
	   !strcmp(el,"dd"))
	   )
    OutString("<%s>", el);

  /* Images and their attributes */
  else if (CharThroughPut && !strcmp(el,"img")) {
    if (EnableGraphics) {
      OutString("<img");
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "src"))
	  OutString(" src=\"%s%s\"", GraphicsVirtualPath, attr[i+1]);
	if (!strcmp(attr[i], "alt")) {
	  if (ProduceXHTML) {
	    OutString(" alt=\"%s\"", attr[i+1]);
	  } else {
	    OutString(" alt=\"");
	    textHandler(NULL, attr[i+1], strlen(attr[i+1]));
	    OutString("\"");
	  }
	}
      }
      OutString(" border=\"0\" />\n");
    } else {
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "alt")) {
	  OutString("%s: %s\n", cGetLImage(cLangCode), attr[i+1]);
	}
      }
    }
  }

  /* Links! Internal and external */
  else if (CharThroughPut && !strcmp(el,"a")) {
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "href")) {
	if (attr[i+1][0] == '#')
	  OutString("<a href=\"%s\">", attr[i+1]);
	else
	  if (Header)
	    /*
	      Might look magic: my documents have URI:s like
	      /article/articleid/printable which means to
	      refer to a non-printable version it's ../articleid
	    */
	    OutString("<a href=\"../%s\">", attr[i+1]);
	  else
	    /* OK this is an ugly fix as of now. Has to be fixed. Commandline parameters or whatever. */
	    OutString("<a href=\"%s\">", attr[i+1]);
      }
      if (!strcmp(attr[i], "name"))
	OutString("<a name=\"%s\">", attr[i+1]);
    }
  }

  /* Footnotes */
  else if (CharThroughPut && !strcmp(el,"footnote")) {
    if (!bIsBook || (!ExtractToc || (ExtractToc && ExtractedToc))) {
      FootnoteNo++;
      OutString("<a href=\"#%d\"><font size=\"-1\"><sup>%d</sup></font></a>", FootnoteNo, FootnoteNo);
      SetStackMode();
      /* If this is the first footnote, create a footnote stacking file */
      if (FootnoteNo == 1) {
	FootnoteStackFile = tmpfile();
	if (FootnoteStackFile == NULL) {
	  fprintf(stderr, "Could not create footnote stack!\n");
	  exit(-1);
	}
      }
      OutString("<p><a name=\"%d\"><font size=\"-1\"><sup>%d</sup></font></a>",
		FootnoteNo,
		FootnoteNo);
    }
    else {
      /* In this situation we could be feeding footnote
	 text to the ToC which must be avoided */
      IgnoreFootnotes = true;
      oldCharThroughPut = CharThroughPut;
      CharThroughPut = false;
    }
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)

{
  char c;

  if (!strcmp(el,"article") ||
      !strcmp(el,"book")) {

    PreserveWhitespace = true;
    CharThroughPut = true;

    /* End ToCs */
    if (bIsBook && ExtractToc) {
      ExtractToc = false;
      ExtractedToc = true;
      /* We don't want another header */
      Header = false;
      if (h3Count) {
	OutString("</ul>\n");
	h3Count = 0;
      }
      if (h2Count) {
	OutString("</ul>\n");
	h2Count = 0;
      }
      if (h1Count) {
	OutString("</ul>\n");
	h1Count = 0;
      }
    }

    /* Add footnotes at the end */
    if (FootnoteNo > 0) {
      if (ProduceXHTML)
	OutString("<hr />\n");
      else
	OutString("<hr>\n");
      OutString("<h2>%s:</h2>\n", cGetLFootnotes(cLangCode));
      if (fseek(FootnoteStackFile, 0, SEEK_SET)) {
	fprintf(stderr, "Could not read temporary footnote stack file.\n");
	exit(-1);
      }
      c = fgetc(FootnoteStackFile);
      while (c != EOF) {
	OutChar(c);
	c = fgetc(FootnoteStackFile);
      }
      fclose(FootnoteStackFile);
    }

    /* 
       Char throughput ends here 
       if we have a headered mode and are running the last iteration,
       OR if we have an extracted ToC but unset flag, write a file
       end.
     */
    if ((Header && !XMLIterations) ||
	(bIsBook && !ExtractToc && ExtractedToc && !XMLIterations)
	)
      OutString("</body>\n</html>\n");
    PreserveWhitespace = false;
    CharThroughPut = false;
  }
  else if (!strcmp(el,"metadata") && Header) {
    OutString("</head>\n");
    CharThroughPut = false;
  }
  else if (!strcmp(el,"body") ||
	   !strcmp(el,"chapter")) {
    PreserveWhitespace = false;
    CharThroughPut = false;
  }
  else if (Header &&
	   CharThroughPut &&
	   !strcmp(el,"description")) {
    PreserveWhitespace = false;
    OutString("</title>\n");
  }
  else if (Header &&
	   CharThroughPut &&
	   (!strcmp(el,"author") ||
	   !strcmp(el,"keywords"))
	   ) {
    PreserveWhitespace = false;
    if (ProduceXHTML)
      OutString("\" />\n");
    else
      /* Problem here */
      OutString("AJ!\">\n");
  }
  else if (!strcmp(el,"published") && Header) {
    ParseDate = false;
    OutString(" Yelah\"");
    OutString(ProduceXHTML ? " />\n" : ">\n");
  }
  else if (!strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")) {
    if (bIsBook && ExtractToc) {
      if (HasNameAttrib)
	OutString("</a>");
      OutString("</li>\n");
      PreserveWhitespace = false;
      CharThroughPut = false;
    } else if (CharThroughPut) {
      if (HasNameAttrib)
	OutString("</a></%s>\n", el);
      else
	OutString("</%s>\n", el);
      LastChar = ' ';
    }
  }
  /* Elements to be plain copied and ended WITH ENTER after them   */
  /* None of the following can be found inside the header elements */
  /* so checking for the Header boolean to be true is unnecessary  */
  else if (
	   CharThroughPut && (
	    !strcmp(el,"p") ||
	    !strcmp(el,"blockquote") ||
	    !strcmp(el,"table") || 
	    !strcmp(el,"tr") ||
	    !strcmp(el,"td") ||
	    !strcmp(el,"ol") ||
	    !strcmp(el,"ul") ||
	    !strcmp(el,"dl") ||
	    !strcmp(el,"li") ||
	    !strcmp(el,"dt") ||
	    !strcmp(el,"dd")
	    )
	   ) {
    OutString("</%s>\n", el);
    LastChar = ' ';
    }
  /* Elements to be plain copied and ended WITHOUT ANY ENTER after them */
  else if (
	   CharThroughPut && (
	   !strcmp(el,"b") ||
	   !strcmp(el,"i") ||
	   !strcmp(el,"a"))
	   )
    OutString("</%s>", el);
  else if (!strcmp(el,"footnote")) {
    if (IgnoreFootnotes) {
      IgnoreFootnotes = false;
      CharThroughPut = oldCharThroughPut;
    } else if (CharThroughPut) {
      OutString("</p>\n");
      UnSetStackMode();
    }
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  register unsigned int i,j;
  unsigned int ucodesize;
  unsigned char tempChar;
  unsigned char plotchar;
  static char temp_string_1[15];
  char *temp_string_2;

  if (!CharThroughPut) {
    /* Do nothing of the throughput hasnt started */
  } else if (ParseDate && (length == 12)) {
  /* A date! - and the length has to be 12 chars */
  } else if (ParseDate && (length == 12)) {
    for (i=0;i<4;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=4;i<6;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=6;i<8;i++)
      OutChar(text[i]);
    OutString(" ");
    for (i=8;i<10;i++)
      OutChar(text[i]);
    OutString(":");
    for (i=10;i<12;i++)
      OutChar(text[i]);
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';

      /* Unicode UTF-8 handling */
      /* XHTML accepts UTF-8 */
      if ((plotchar > 0x7F)  && (!ProduceXHTML)) {
	tempChar = plotchar;
	ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	for (j=0;j<ucodesize;j++) {
	  temp_string_1[j] = text[i];
	  i++;
	}
	i--;
	temp_string_1[j] = '\0';
	/* Create recoding engine and request if needed */
	if (recode_engine_outer == NULL)
	  recode_engine_outer = recode_new_outer(true);
	if (recode_engine_request_2 == NULL) {
	  recode_engine_request_2 = recode_new_request(recode_engine_outer);
	  recode_scan_request(recode_engine_request_2, "utf8..h4");
	}
	temp_string_2 = recode_string(recode_engine_request_2, temp_string_1);
	if (temp_string_2) {
	  /* else just mash out the raw entity ... */
	  if (CharThroughPut)
	    OutString(temp_string_2);
	  free(temp_string_2);
	  LastChar = '\0';
	  continue;
	} else {
	  plotchar = ' ';
	}
      }

      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  OutChar(plotchar);
      } else {
	/* Just mash out the text ... */
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' ')) {
	  /* Some HTML characters require special treatment */
	  if (plotchar == '<')
	    OutString("&lt;");
	  else if (plotchar == '>')
	    OutString("&gt;");
	  else if (plotchar == '&')
	    OutString("&amp;");
	  else
	    OutChar(plotchar);
	}
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}

/* All raw entities are handled here                   */
/* a true entity handler / cross reference to be added */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  int i;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;

  /* The article has to start first, and that we want to output chars */
  if (CharThroughPut) {
    if (ProduceXHTML) {
      if (s[0]=='&' && s[length-1]==';') {
	/* Create recoding engine and request if needed */
	if (recode_engine_outer == NULL)
	  recode_engine_outer = recode_new_outer(true);
	if (recode_engine_request_1 == NULL) {
	  recode_engine_request_1 = recode_new_request(recode_engine_outer);
	  recode_scan_request(recode_engine_request_1, "h4..utf8");
	}
	for(i=0;i<length;i++)
	  temp_string_1[i]=s[i];
	temp_string_1[length] = '\0';
	temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
	if (temp_string_2)
	  OutString(temp_string_2);
	free(temp_string_2);
      }
    } else {
      /* Just mash out the raw entity ... */
      for (i=0;i<length;i++)
	OutChar(s[i]);
      /* We don't know for sure the corresponding ISO8859-1 character value */
      LastChar = '\0';
    }
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  if (!strcmp(argv[arg],"-noheader"))
    Header = false;
  if (!strcmp(argv[arg],"-toc")) {
    ExtractToc = true;
    fprintf(stderr, "Including Table of Contents in output file\n");
  }
  if (!strcmp(argv[arg],"-xhtml"))
    ProduceXHTML = true;
  if (!strcmp(argv[arg],"-gphx"))
    EnableGraphics = true;
  if (!strcmp(argv[arg],"-gppv")) {
    if(*argv[arg+1] == '-') {
      fprintf(stderr, "You must give a <path> parameter to the -gppv switch!\n");
      exit(-1);
    }
    strncpy(GraphicsVirtualPath, argv[arg+1], 127);
    GraphicsVirtualPath[127] = '\0';
    if (GraphicsVirtualPath[strlen(GraphicsVirtualPath)-1] != '/')
      strcpy(GraphicsVirtualPath+strlen(GraphicsVirtualPath), "/");
    fprintf(stderr, "Using virtual path: %s\n", GraphicsVirtualPath);
  }
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  printf("         -gphx         = Enables graphics inclusion\n");
  printf("         -gppv <path>  = Virtual path to graphics (-gpath /images/)\n");
  printf("         -noheader     = Write no headers\n");
  printf("         -xhtml        = Produce XHTML 1.0\n");
}

/* End exported functions */
