/* xmltopdb.c
   Program for parsing and converting article.dtd compliant 
   XML files to Aportis Doc or RichReader for Palm Pilot formats
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the Yelah xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "common.h"
#include <pilotpdb.h>

/* Language identifier "sv"=Swedish, "en"=English etc... */
char *cLangCode;
/* Parse a date */
bool ParseDate = false;
/* Reading description to string? */
bool ParseDescription = false;
char cDescription[32] = "";
/* Has the article begun at all? */
bool bTextThroughput = false;
/* Are we working with an unnumbered (bulleted) table? */
bool UnnumberedMode = false;
int NumberedMode = false;
/* This one signals rich format of text */
bool RichText = false;
/* This one tells us if Palm Pilot output should be compressed */
bool Compressed = true;
/* 
   This one counts paragraphs, used for identifying blockquotes
   with inlined paragraphs.
*/
int ParaCounter = 0;
/* This one to remove spaces etc */
static unsigned char LastChar = ' ';

/* Temporary file to buffer all output */
FILE * BufferedOutput = NULL;

/****************************************************************
 * The converter also supports the RichReader extensions for
 * the RichReader if you want to use them. Invoking the program
 * with the -rich flag sets the RichText boolean to true thus
 * enabling the RichReader extensions. The RichReader extensions
 * used here have been extracted by example from documents
 * created by RichReader and is not complete. The RichReader 
 * format adds these characters:
 *
 * Each command begins with 0x7f followed by one of the following
 * commands which are the OR:ed (|) together:
 *
 * 0x00 No special text formatting (first bit zero)
 * 0x01 ?
 * 0x02 Italic
 * 0x04 Bold
 * 0x20 Small font
 * 0x40 Normal font
 * 0X60 Large font
 *
 * These commands cannot be OR:ed into the other ones (font size
 * etc) but stand on their own:
 *
 * 0x80 Left justification
 * 0x81 Right Justification
 * 0x82 Both left and right justification
 * 0x83 Centered justification
 * 0x8c Horizontal ruler
 * 0xc0 No indentation (reset)
 * 0xc1 indent one char
 * 0xc2 indent two chars
 * 0xc3 indent three chars ... and so forth
 * 0xfe Turn on "indentation is at first tab stop"
 * 0xf0 Turn off "indentation is at first tab stop"
 ****************************************************************/

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  int i;

  /* Get language tag */
  if (!strcmp(el,"metadata")) {
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "language"))
	cLangCode = cFixLangCode(attr[i+1]);
    }
  }
  /* The only metadata extracted for pdb documents is description */
  else if (!strcmp(el,"description")) {
    bTextThroughput = true;
    PreserveWhitespace = true;
    ParseDescription = true;
  }
  else if (!strcmp(el,"body") ||
	   !strcmp(el,"chapter")) {
    bTextThroughput = true;
    /* No line breaking. Uneccessary really - we don't use OutString or OutChar */
    LineBreakMode = false;
    if (BufferedOutput == NULL)
      BufferedOutput = tmpfile();
  }
  /* Elements to be copied with a justify attribute */
  else if (!strcmp(el,"h1")) {
    PreserveWhitespace = true;
    if (RichText) {
      if (!attr[0])
	fprintf(BufferedOutput, "\x7f\x64");
      else for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "align")) {
	  if (!strcmp(attr[i+1], "right"))
	    fprintf(BufferedOutput, "\x7f\x81\x7f\x64");
	  else if (!strcmp(attr[i+1], "center"))
	    fprintf(BufferedOutput, "\x7f\x83\x7f\x64");
	  else if (!strcmp(attr[i+1], "justify"))
	    fprintf(BufferedOutput, "\x7f\x82\x7f\x64");
	  else
	    fprintf(BufferedOutput, "\x7f\x80\x7f\x64");
	}
      }
    }
  }
  else if (!strcmp(el,"h2") || !strcmp(el,"h3")) {
    PreserveWhitespace = true;
    if (RichText) {
      if (!attr[0])
	fprintf(BufferedOutput, "\x7f\x60");
      else for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "align")) {
	  if (!strcmp(attr[i+1], "right"))
	    fprintf(BufferedOutput, "\x7f\x81\x7f\x60");
	  else if (!strcmp(attr[i+1], "center"))
	    fprintf(BufferedOutput, "\x7f\x83\x7f\x60");
	  else if (!strcmp(attr[i+1], "justify"))
	    fprintf(BufferedOutput, "\x7f\x82\x7f\x60");
	  else
	    fprintf(BufferedOutput, "\x7f\x80\x7f\x60");
	}
      }
    }
  }
  /* Paragraph */
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = true;
    ParaCounter++;
    if (RichText && attr[0]) {
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "align")) {
	  if (!strcmp(attr[i+1], "right"))
	    fprintf(BufferedOutput, "\x7f\x81");
	  else if (!strcmp(attr[i+1], "center"))
	    fprintf(BufferedOutput, "\x7f\x83");
	  else if (!strcmp(attr[i+1], "justify"))
	    fprintf(BufferedOutput, "\x7f\x82");
	  else
	    fprintf(BufferedOutput, "\x7f\x80");
	}
      }
    }
  }
  else if (!strcmp(el,"br"))
    fprintf(BufferedOutput, "\n");
  /* Horizontal rulers */
  else if (!strcmp(el,"hr")) {
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x8c\n");
    else
      fprintf(BufferedOutput, "\n\n--- * ---\n\n");
  }
  else if (!strcmp(el, "blockquote")) {
    /* The blockquote is treated like a special kind of indented paragraph */
    /* Have to do something about this in the charmongler routine...       */
    PreserveWhitespace = true;
    ParaCounter = 0;
    if (RichText)
      fprintf(BufferedOutput, "\x7f\xc3");
  }
  /* This part handles different kinds of lists... */
  else if (!strcmp(el, "ul")) {
    UnnumberedMode = true;
  } else if (!strcmp(el, "ol")) {
    /* Yeah use the flag as list enumerator... hard 2 c? */
    NumberedMode++;
  } else if (!strcmp(el, "li")) {
    PreserveWhitespace = true;
    if (UnnumberedMode)
      if (RichText)
	fprintf(BufferedOutput, "\x7f\xfe\x7f\xc2\x95\x09");
      else
	fprintf(BufferedOutput, "*  ");
    else if (NumberedMode) {
      if (RichText)
	fprintf(BufferedOutput, "\x7f\xfe\x7f\xc2%d.\x09", NumberedMode);
      else
	fprintf(BufferedOutput, "%d. ", NumberedMode);
      NumberedMode++;
    }
  }
  else if (!strcmp(el, "dl")) {
    /* Do nothing */
  } else if (!strcmp(el, "dt")) {
    PreserveWhitespace = true;
  } else if (!strcmp(el, "dd")) {
    PreserveWhitespace = true;
    fprintf(BufferedOutput, "        ");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
    /* Do nothing */
  } else if (!strcmp(el,"tr")) {
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = true;
  }
  /* Text formatting */
  else if (!strcmp(el,"b")) {
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x44");
    else
      fprintf(BufferedOutput, "_");
  }
  else if (!strcmp(el,"i")) {
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x42");
    else
      fprintf(BufferedOutput, "*");
  }
  /* Images and their attributes */
  else if (!strcmp(el,"img")) {
    fprintf(BufferedOutput, "<Image: ");
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "alt"))
	fprintf(BufferedOutput, attr[i+1]);
    }
    fprintf(BufferedOutput, ">\n");
  }
  /* Footnotes */
  else if (!strcmp(el,"footnote")) {
    FootnoteNo++;
    fprintf(BufferedOutput, "(%d)", FootnoteNo);
    StackMode = true;
    /* If this is the first footnote, create a footnote stacking file */
    if (FootnoteNo == 1)
      FootnoteStackFile = tmpfile();
    fprintf(FootnoteStackFile, "(%d) ", FootnoteNo);
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  WORD i;
  /* List for storing the Pilot File */
  tPilotRecord *PilotRecords = NULL;
  /* Used for shoveling data chunks */
  char c;
  byte byteBuffer[MAGIC_CHUNKSIZE];

  if (
      !strcmp(el,"article") ||
      !strcmp(el,"book")) {

    bTextThroughput = false;

    /* Stuff up the footnotes at the end of the file */
    if (FootnoteNo > 0) {
      fprintf(BufferedOutput, "\n\n\n%s:\n\n", cGetLFootnotes(cLangCode));
      if (fseek(FootnoteStackFile, 0, SEEK_SET)) {
	fprintf(stderr, "Could not read temporary footnote stack file.\n");
	exit(-1);
      }
      c = fgetc(FootnoteStackFile);
      while (c != EOF) {
	fputc(c, BufferedOutput);
	c = fgetc(FootnoteStackFile);
      }
      fclose(FootnoteStackFile);
    }
    PreserveWhitespace = false;
    fprintf(BufferedOutput, "\n");

    /* Chunkize the buffered file */
    if (fseek(BufferedOutput, 0, SEEK_SET)) {
      fprintf(stderr, "Could not read temporary output buffer file.\n");
      exit(-1);
    }

    /*****************************************************************
     * The record attribute for aportis doc is obviously "dirty".
     *****************************************************************/
    SetPDBFileType("TEXt", "REAd", RECATT_DIRTY);
    i = 0;
    c = fgetc(BufferedOutput);
    while (c != EOF) {
      byteBuffer[i] = c;
      i++;
      if (i == MAGIC_CHUNKSIZE) {
	PilotRecords = AddRecord(PilotRecords, byteBuffer, MAGIC_CHUNKSIZE, (DWORD) 0);
	i = 0;
      }
      c = fgetc(BufferedOutput);
    }
    PilotRecords = AddRecord(PilotRecords, byteBuffer, i, (DWORD) 0);
    if (Compressed)
      DocCompressList(PilotRecords);
    fclose(BufferedOutput);
    WritePDBFile(PilotRecords, OutFile, cDescription);
    FreeRecords(PilotRecords);
    /* 
       After this point we will not write anything more to the file.
       It will be closed by the main parsin commandline program.
    */
  }
  /* Only description to be parsed */
  else if (!strcmp(el,"description")) {
    bTextThroughput = false;
    PreserveWhitespace = false;
    ParseDescription = false;
  }
  /* Just flowing paragraphs here */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")) {
    PreserveWhitespace = false;
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x40\x7f\x80\n\n");
    else
      fprintf(BufferedOutput, "\n\n");
  }
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = false;
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x80\n\n");
    else
      fprintf(BufferedOutput, "\n\n");
  }
  else if (!strcmp(el,"blockquote")) {
    PreserveWhitespace = false;
    if (!ParaCounter)
      fprintf(BufferedOutput, "\n\n");
    if (RichText)
      fprintf(BufferedOutput, "\x7f\xc0");
  }
  /* This section handles different kinds of lists... */
  else if (!strcmp(el,"ul")) {
    UnnumberedMode = false;
    fprintf(BufferedOutput, "\n");
  } else if (!strcmp(el,"ol")) {
    NumberedMode = 0;
    fprintf(BufferedOutput, "\n");
  } else if (!strcmp(el,"dl")) {
    fprintf(BufferedOutput, "\n");
  } else if (!strcmp(el,"li") ||
	     !strcmp(el,"dt") ||
	     !strcmp(el,"dd")
	     ) {
    PreserveWhitespace = false;
    /* You know they all look the same in the end */
    if (RichText)
      fprintf(BufferedOutput, "\x7f\xf0\x7f\xc0\n");
    else
      fprintf(BufferedOutput, "\n\n");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
  } else if (!strcmp(el,"tr")) {
    fprintf(BufferedOutput, "\n");
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = false;
    fprintf(BufferedOutput, "\t");
  }
  /* Elements to be plain copied and ended WITHOUT ANY ENTER after them */
  else if (!strcmp(el,"b")) {
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x40");
    else
      fprintf(BufferedOutput, "_");
  }
  else if (!strcmp(el,"i")) {
    if (RichText)
      fprintf(BufferedOutput, "\x7f\x40");
    else
      fprintf(BufferedOutput, "*");
  }
  else if (!strcmp(el,"footnote")) {
    StackMode = false;
    PreserveWhitespace = true;
    fprintf(FootnoteStackFile, "\n\n");
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  register unsigned int i, j;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static char temp_string_1[15];
  char *temp_string_2;

  if (!bTextThroughput) {
    /* Do nothing of the Article hasnt started */
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';

      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	for (j=0;j<ucodesize;j++) {
	  temp_string_1[j] = text[i];
	  i++;
	}
	i--;
	temp_string_1[j] = '\0';
	/* Create recoding engine and request if needed */
	if (recode_engine_outer == NULL)
	  recode_engine_outer = recode_new_outer(true);
	if (recode_engine_request_2 == NULL) {
	  recode_engine_request_2 = recode_new_request(recode_engine_outer);
	  recode_scan_request(recode_engine_request_2, "utf8..latin1");
	}
	temp_string_2 = recode_string(recode_engine_request_2, temp_string_1);
	if (temp_string_2) {
	  plotchar = temp_string_2[0];
	  free(temp_string_2);
	} else {
	  plotchar = ' ';
	}
      }

      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  fputc(plotchar, BufferedOutput);
      } else {
	/* Just mash out the text ... */
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' ')) {
	  if (ParseDescription) {
	    if (strlen(cDescription) < 31) {
	      j = strlen(cDescription);
	      cDescription[j] = plotchar;
	      cDescription[j+1] = '\0';
	    }
	  } else {
	    if (!StackMode)
	      fputc(plotchar, BufferedOutput);
	    else
	      fputc(plotchar, FootnoteStackFile);
	  }
	}
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}


/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  register int i,j;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;
  
  /* Only valid escapes */
  if (s[0]=='&' && s[length-1]==';') {
    /* Create recoding engine and request if needed */
    if (recode_engine_outer == NULL)
      recode_engine_outer = recode_new_outer(true);
    if (recode_engine_request_1 == NULL) {
      recode_engine_request_1 = recode_new_request(recode_engine_outer);
      recode_scan_request(recode_engine_request_1, "h4..latin1");
    }
    for(i=0;i<length;i++)
      temp_string_1[i]=s[i];
    temp_string_1[length] = '\0';
    temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
    if (temp_string_2) {
      /* else just mash out the raw entity ... */
      if (ParseDescription) {
	for (i=0;i<strlen(temp_string_2);i++) {
	  if (strlen(cDescription) < 31) {
	    j = strlen(cDescription);
	    cDescription[j] = temp_string_2[i];
	    cDescription[j+1] = '\0';
	    LastChar = temp_string_2[i];
	  }
	}
      } else if (!StackMode && bTextThroughput) {
	if (BufferedOutput == NULL)
	  BufferedOutput = tmpfile();
	fprintf(BufferedOutput, "%s", temp_string_2);
      } else if (StackMode && bTextThroughput)
	fprintf(FootnoteStackFile, "%s", temp_string_2);
      free(temp_string_2);
    }
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  if (!strcmp(argv[arg],"-rich"))
    RichText = true;
  if (!strcmp(argv[arg],"-nopack"))
    Compressed = false;
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  printf("         -rich         = Write rich formatting code (Palm Pilot)\n");
  printf("         -nopack       = Do not compress output (Palm Pilot)\n");
}

/* End exported functions */
