/* xmltortf.c
   Program for parsing and converting article.dtd compliant 
   XML files to Microsoft(TM) RTF (Rich Text Format)
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include <stdarg.h>
#include "common.h"

/* Language identifier "sv"=Swedish, "en"=English etc... */
char *cLangCode;
/* Parse a date */
bool ParseDate = false;
/* Has the article begun at all? */
bool bTextThroughput = false;
/* Are we working with an unnumbered (bulleted) table? */
bool UnnumberedMode = false;
int NumberedMode = 0;
/*
   This one counts paragraphs, used for identifying blockquotes
   with inlined paragraphs.
*/
int ParaCounter = 0;
/* Counting chapters in books */
int ChapCounter = 0;

/*
  ROUTINES
  Building tables in memory as linked 2D matrix list
*/

typedef struct tRow tRow;
typedef struct tColumn tColumn;

struct tRow {
  tColumn *columns;
  tRow *next;
};

struct tColumn {
  char *cell;
  tColumn *next;
};

/* Global variables used for storing tables */
tRow *Table;
char *MemCell;
int bWriteCell = false;

static tRow *
newRow()
{
  tRow *newp;
  
  newp = (tRow *) malloc(sizeof(tRow));
  newp->columns = NULL;
  newp->next = NULL;
  return newp;
}

static tColumn *
newColumn(char *data)
{
  tColumn *newp;

  newp = (tColumn *) malloc(sizeof(tColumn));
  newp->cell = data;
  newp->next = NULL;
  return newp;
}

static tRow *
AddRow(tRow *table)
{
  tRow *p;

  if (table == NULL)
    return newRow();
  for (p = table; p->next != NULL; p = p->next)
    ;
  p->next = newRow();
  return table;
}

static void 
AddCell(tRow *table, char *data)
{
  tRow *p;
  tColumn *q;

  /* This first thing should actually never happen */
  if (table == NULL)
    p = newRow();
  /* But this is the normal operation */
  else
    for (p = table; p->next != NULL; p = p->next)
      ;
  if (p->columns == NULL) {
    p->columns = newColumn(data);
  } else {
    for (q = p->columns; q->next != NULL; q = q->next)
      ;
    q->next = newColumn(data);
  }
}

static void 
DeallocateTable(tRow *table)
{
  tRow *p, *r;
  tColumn *q, *s;
  
  if (table == NULL)
    return;
  p = table;
  while (p != NULL) {
    q = p->columns;
    while (q != NULL) {
      if (q->cell != NULL)
	free(q->cell);
      s = q->next;
      free(q);
      q = s;
    }
    r = p->next;
    free(p);
    p = r;
  }
}

/* Returns the max width (in columns) of a table */
static int 
GetCols(tRow *table)
{
  tRow *p;
  tColumn *q;
  register int tmp, cols;
  
  /* Could be dangerous to return 0 here... */
  if (table == NULL)
    return 1;
  p = table;
  cols = 0;
  while (p != NULL) {
    q = p->columns;
    tmp = 0;
    while (q != NULL) {
      tmp++;
      q = q->next;
    }
    if (tmp>cols)
      cols = tmp;
    p = p->next;
  }
  if (cols == 0)
    return (1);
  else
    return (cols);
}

static void 
NormalizeTable(tRow *table)
{
  tRow *p;
  tColumn *q;
  char *EmptyString;
  register int cols, tmp;

  if (table == NULL)
    return;
  cols = GetCols(table);
  p = table;

  while (p != NULL) {
    q = p->columns;
    tmp = 0;
    while (q != NULL) {
      tmp++;
      q = q->next;
    }
    q = p->columns;
    while(q->next != NULL)
      q = q->next;
    while (tmp<cols) {
      /*
	Add empty columns until we have
	equal width of all rows, some may
	need a space character in them.
      */
      EmptyString = (char *) malloc(2 * sizeof(char));
      EmptyString[0] = ' ';
      EmptyString[1] = '\0';
      q->next = newColumn(EmptyString);
      q = q->next;
      tmp ++;
    }
    p = p->next;
  }
}

static void 
PlotTable(tRow *table)
{
  tRow *p;
  tColumn *q;
  int i, cols, delta;

  OutString("\\par\n{\n");
  if (table == NULL)
    return;
  p = table;
  while (p != NULL) {
    OutString("\\trowd\n");
    /*
      Plot cell sizes 8747 / cells
      8747 is the width of the A4 paper I think
      so perhaps this number should be lower...
     */
    cols = GetCols(table);
    delta = 8747 / cols;
    for (i=0; i<cols; i++) {
      OutString("\\cellx%d ", delta*(i+1));
    }
    OutString("\n\\intbl\n");
    q = p->columns;
    while (q != NULL) {
      OutString("%s\\cell\n", q->cell);
      q = q->next;
    }
    OutString("\\row\n");
    p = p->next;
  }
  OutString("}\n");
}

static char *
OutMemS(char *AppendString, char *String)
{
  if (AppendString == NULL) {
    AppendString = malloc((strlen(String)+1) * sizeof(char));
    strcpy(AppendString, String);
  } else {
    AppendString = (char *) realloc(AppendString, (strlen(String) + strlen(AppendString) + 1) * sizeof(char) );
    strcat(AppendString, String);
  }
  return AppendString;
}

static char *
OutMemC(char *AppendString, char c)
{
  int i;
  
  if (AppendString == NULL) {
    AppendString = (char *) malloc(2 * sizeof(char));
    i = 0;
  } else {
    AppendString = (char *) realloc(AppendString, (strlen(AppendString) + 2) * sizeof(char));
    i = strlen(AppendString);
  }
  AppendString[i] = c;
  AppendString[i+1] = '\0';
  return AppendString;
}

void 
OutStringCellCheck(char *fmt, ...)
{
  char tmpBuffer[160];
  va_list ap;

  va_start(ap, fmt);
  if (!bWriteCell) {
    /* This routine just passes the string through */
    OutString(fmt, ap);
  } else {
    vsprintf(tmpBuffer, fmt, ap);
    MemCell = OutMemS(MemCell, tmpBuffer);
  }
  va_end(ap);
}

void 
OutCharCellCheck(char c)
{
  if (!bWriteCell)
    OutChar(c);
  else
    MemCell = OutMemC(MemCell, c);
}

/*
  Routine to translate ISO8859-1 to escape-sequences
*/

static char *
Escape(unsigned char IsoChar)
{
  char* token;
  char* tmpointer;
  /* This holds the number of elements in the array */
  static int ComparrayElements = 0;
  /* The following line limits number of entities to 256 */
  static char* Comparray[128];
  static char Entline[STR_MAXLEN] = "";
  FILE* ISOentities;

  /* Build an ISO Latin-1 array from file if the array == NULL */
  if (!ComparrayElements) {
    /* Build our dynamic array */
    /* This way of finding the shared files is probably
       not 100% accurate. Will try to do better or get
       the RTF patch into recode so this isn't needed... */
    ISOentities = fopen("iso-rtf.tbl", "r");
    if (ISOentities == NULL)
      ISOentities = fopen("/usr/share/" PACKAGE "/iso-rtf.tbl", "r");
    if (ISOentities == NULL)
      ISOentities = fopen("/usr/local/share/" PACKAGE "/iso-rtf.tbl", "r");
    if (ISOentities == NULL)
      ISOentities = fopen("./share/" PACKAGE "/iso-rtf.tbl", "r");
    if (ISOentities == NULL) {
      fprintf(stderr, "Could not read the parameter file iso-rtf.tbl!\n");
      exit(-1);
    }
    while(!(fgets(Entline, STR_MAXLEN, ISOentities) == NULL)) {
      if(!(Entline[0]=='#') && !(Entline[0]=='\n')) {
	token = Entline;
	if (token != NULL) {
	  /* "token" now contains the entity like &auml; or so */
	  /* Remove any CRLF stuff at the end */
	  while((unsigned char) token[strlen(token)-1] < (unsigned char) ' ') {
	    token[strlen(token)-1] = '\0';
	  }
	  tmpointer = (char*) malloc(strlen(token));
	  strcpy(tmpointer, token);
	  Comparray[ComparrayElements] = tmpointer;
	  ComparrayElements++;
	}
      }
    }
    fclose(ISOentities);
  }

  /* And this code actually locates the entity in the array */
  return (Comparray[IsoChar-0xA0]);
}


/* Routine to translate escaped entities to RTF macros */
static char *
Translate(char* Entity)
{
  int i = 0;
  bool done = false;
  char* token;
  char* tmpointer;

  /* This holds the number of elements in the array */
  static int ComparrayElements = 0;
  /* The following line limits number of entities to 256 */
  static char* Comparray[256];
  static char Entline[STR_MAXLEN] = "";
  FILE* RTFentities;
  
  /* Build an ISO Latin-1 array from file if the array == NULL */
  if (!ComparrayElements) {
    /* Build our dynamic array */
    /* This way of finding the shared files is probably
       not 100% accurate. Will try to do better or get
       the RTF patch into recode so this isn't needed... */
    RTFentities = fopen("lat1-rtf.tbl", "r");
    if (RTFentities == NULL)
      RTFentities = fopen("/usr/share/" PACKAGE "/lat1-rtf.tbl", "r");
    if (RTFentities == NULL)
      RTFentities = fopen("/usr/local/share/" PACKAGE "/lat1-rtf.tbl", "r");
    if (RTFentities == NULL)
      RTFentities = fopen("./share/" PACKAGE "/lat1-rtf.tbl", "r");
    if (RTFentities == NULL) {
      fprintf(stderr, "Could not read the parameter file lat1-rtf.tbl!\n");
      exit(-1);
    }
    while(!(fgets(Entline, STR_MAXLEN, RTFentities) == NULL)) {
      if(!(Entline[0]=='#')) {
	token = (char*) strtok(Entline,"\t");
	if (token != NULL) {
	  /* "token" now contains the entity like auml or so */
	  tmpointer = (char*) malloc(strlen(token));
	  strcpy(tmpointer, token);
	  Comparray[ComparrayElements] = tmpointer;
	  ComparrayElements++;
	  token = (char*) strtok(NULL,"\t");
	  /* This line removes crriage returns etc from RTF entity */
	  while(token[strlen(token)-1] < ' ')
	    token[strlen(token)-1]='\0';
	  tmpointer = (char*) malloc(strlen(token));
	  strcpy(tmpointer, token);
	  Comparray[ComparrayElements] = tmpointer;
	  ComparrayElements++;
	}
      }
    }
    fclose(RTFentities);
  }

  /* And this code actually locates the entity in the array */
  while (!done && i<ComparrayElements) {
    if (!strcmp(Comparray[i],Entity)) {
      done = true;
      return (Comparray[i+1]);
    }
    i+=2;
  }
  return(NULL);
}


/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  register int i;
  int bFoundAlignment;

  if (!strcmp(el,"article") ||
      !strcmp(el,"book")) {
    bTextThroughput = true;
    OutString("{\\rtf1\\ansi \\deff0{\\fonttbl{\\f16\\froman Times New Roman;}} ");
  }
  else if (!strcmp(el,"metadata")) {
    OutString("{\\info");
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "cathegory"))
	OutString("{\\subject %s}", attr[i+1]);

      /* Handle the language somehow... */
      if (!strcmp(attr[i], "language")) {
	cLangCode = cFixLangCode(attr[i+1]);
	/* OK this is not so cute */
	if (!strcmp(cLangCode,"SV")) {
	  /* Swedish language attribute */
	  OutString("\\deflang1053 ");
	} else {
	  /* US English language code */
	  OutString("\\deflang1033 ");
	}
      }
    }
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = true;
    OutString("{\\title ");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = true;
    OutString("{\\author ");
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = true;
    OutString("{\\keywords ");
  }
  else if (!strcmp(el,"published")) {
    /* Set a flag to the text parser and handle this correctly */
    ParseDate = true;
    OutString("{\\creatim");
  }
  else if (!strcmp(el,"body")) {
    OutString("\\pard\\plain\\f16\\fs20\\sa120 ");
  }
  /* Chapter heading (for books) */
  else if (!strcmp(el,"chapter")) {
    if (ChapCounter)
      OutString("\\page");
    ChapCounter++;
    OutString("\\pard\\plain\\f16\\fs20\\sa120 ");
  }
  /* Elements to be copied with a justify attribute */
  /* First level heading */
  else if (!strcmp(el,"h1")) {
    PreserveWhitespace = true;
    bFoundAlignment = false;
    /* We want new pages for each chapter in a book */
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right"))
	  OutString("\\pard\\plain\\b\\fs40\\qr\\sa120 ");
	else if (!strcmp(attr[i+1], "center"))
	  OutString("\\pard\\plain\\b\\fs40\\qc\\sa120 ");
	else if (!strcmp(attr[i+1], "justify"))
	  OutString("\\pard\\plain\\b\\fs40\\qj\\sa120 ");
	else
	  OutString("\\pard\\plain\\b\\fs40\\ql\\sa120 ");
	bFoundAlignment = true;
      }
    }
    if (!bFoundAlignment)
      OutString("\\pard\\plain\\b\\fs40\\ql\\sa120 ");
  }
  /* Second level heading */
  else if (!strcmp(el,"h2")) {
    PreserveWhitespace = true;
    bFoundAlignment = false;
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right"))
	  OutString("\\pard\\plain\\b\\fs30\\qr\\sa120 ");
	else if (!strcmp(attr[i+1], "center"))
	  OutString("\\pard\\plain\\b\\fs30\\qc\\sa120 ");
	else if (!strcmp(attr[i+1], "justify"))
	  OutString("\\pard\\plain\\b\\fs30\\qj\\sa120 ");
	else
	  OutString("\\pard\\plain\\b\\fs30\\ql\\sa120 ");
	bFoundAlignment = true;
      }
    }
    if (!bFoundAlignment)
      OutString("\\pard\\plain\\b\\fs30\\ql\\sa120 ");
  }
  /* Third level heading */
  else if (!strcmp(el,"h3")) {
    PreserveWhitespace = true;
    bFoundAlignment = false;
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right"))
	  OutString("\\pard\\plain\\b\\fs25\\qr\\sa120 ");
	else if (!strcmp(attr[i+1], "center"))
	  OutString("\\pard\\plain\\b\\fs25\\qc\\sa120 ");
	else if (!strcmp(attr[i+1], "justify"))
	  OutString("\\pard\\plain\\b\\fs25\\qj\\sa120 ");
	else
	  OutString("\\pard\\plain\\b\\fs25\\ql\\sa120 ");
	bFoundAlignment = true;
      }
    }
    if (!bFoundAlignment)
      OutString("\\pard\\plain\\b\\fs25\\ql\\sa120 ");
  }
  /* Paragraph */
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = true;
    ParaCounter++;
    if (!attr[0])
      OutString("\\plain\\ql\\sa120 ");
    else for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right"))
	  OutString("\\plain\\qr\\sa120 ");
	else if (!strcmp(attr[i+1], "center"))
	  OutString("\\plain\\qc\\sa120 ");
	else if (!strcmp(attr[i+1], "justify"))
	  OutString("\\plain\\qj\\sa120 ");
	else
	  OutString("\\plain\\ql\\sa120 ");
      }
    }
  }
  else if (!strcmp(el,"br"))
    OutString("\\line ");
  else if (!strcmp(el,"hr"))
    OutString("\\par\n");
  else if (!strcmp(el, "blockquote")) {
    /* The blockquote is treated like a special kind of indented paragraph */
    PreserveWhitespace = true;
    ParaCounter = 0;
    OutString("\\pard\\plain\\ql\\li567\\ri567\\sa120 ");
  }
  /* This part handles different kinds of lists... */
  else if (!strcmp(el, "ul")) {
    UnnumberedMode = true;
    OutString("\\pard\\plain\\ql\\li360\\fi-360\\sa0 ");
  } else if (!strcmp(el, "ol")) {
    /* Yeah use the flag as list enumerator... hard 2 c? */
    NumberedMode = 1;
    OutString("\\pard\\plain\\ql\\li360\\fi-360\\sa0 ");
  } else if (!strcmp(el, "li")) {
    PreserveWhitespace = true;
    if (UnnumberedMode)
      OutString("\\bullet\\tab ");
    else if (NumberedMode) {
      OutString("%d.\\tab ", NumberedMode);
      NumberedMode++;
    }
  }
  else if (!strcmp(el, "dl")) {
    OutString("\\pard\\plain\\ql\\sa0 ");
  } else if (!strcmp(el, "dt")) {
    PreserveWhitespace = true;
    OutString("\\li0 ");
  } else if (!strcmp(el, "dd")) {
    PreserveWhitespace = true;
    OutString("\\li360 ");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
    Table = NULL;
  } else if (!strcmp(el,"tr")) {
    Table = AddRow(Table);
  } else if (!strcmp(el,"td")) {
    MemCell = NULL;
    PreserveWhitespace = true;
    bWriteCell = true;
    /* Flag to send to string buffer */
  }
  /* Text formatting */
  else if (!strcmp(el,"b"))
    OutString("\\b ");
  else if (!strcmp(el,"i"))
    OutString("\\i ");
  /* Images and their attributes */
  else if (!strcmp(el,"img")) {
    OutString("%s: ", cGetLImage(cLangCode));
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "alt"))
	OutString("%s", attr[i+1]);
    }
    OutString("\\par ");
  }

  /* Footnotes */
  else if (!strcmp(el,"footnote")) {
    OutString("\\plain\\fs15\\up6 \\chftn {\\footnote \\pard \\plain\\fs15\\up6 \\chftn \\plain\\sa120 ");
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  if (!strcmp(el,"article") ||
      !strcmp(el,"book")) {
    bTextThroughput = false;
    OutString("}");
  }
  else if (!strcmp(el,"metadata")) {
    OutString("{\\doccomm Document created by XMLTOOLS converter}{\\*\\hlinkbase http://www.df.lth.se/\\'7etriad/krad/}}");
  }
  else if (!strcmp(el,"body")) {
    /* OutString("");  } ? */
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = false;
    OutString("}");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = false;
    OutString("}");
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = false;
    OutString("}");
  }
  else if (!strcmp(el,"published")) {
    ParseDate = false;
    OutString("}");
  }
  /* Just flowing paragraphs here */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")) {
    PreserveWhitespace = false;
    OutString("\\b0\n\\par ");
  }
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = false;
    OutString("\n\\par ");
  }
  else if (!strcmp(el,"blockquote")) {
    PreserveWhitespace = false;
    if (ParaCounter)
      OutString("\n\\li0\\ri0 ");
    else
      OutString("\n\\par\\li0\\ri0 ");
  }
  /* This section handles different kinds of lists... */
  else if (!strcmp(el,"ul") ||
	   !strcmp(el,"ol")) {
    UnnumberedMode = false;
    OutString("\\li0\\fi0\\par ");
  } else if (!strcmp(el,"dl")) {
    OutString("\\li0\\par ");
  } else if (!strcmp(el,"li") ||
	     !strcmp(el,"dt") ||
	     !strcmp(el,"dd")
	     ) {
    PreserveWhitespace = false;
    /* You know they all look the same in the end */
    OutString("\\par ");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
    NormalizeTable(Table);
    PlotTable(Table);
    DeallocateTable(Table);
    OutString("\\par ");
  } else if (!strcmp(el,"tr")) {
    /* Do nothing */
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = false;
    AddCell(Table, MemCell);
    bWriteCell = false;
  }
  /* Elements to be plain copied and ended WITHOUT ANY ENTER after them */
  else if (!strcmp(el,"b"))
    OutString("\\b0 ");
  else if (!strcmp(el,"i"))
    OutString("\\i0 ");
  else if (!strcmp(el,"footnote")) {
    OutString("}\\plain ");
    PreserveWhitespace = true;
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  unsigned int i;
  unsigned int ucode;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static unsigned char LastChar = ' ';

  if (!bTextThroughput) {
    /* Do nothing of the Article hasnt started */
  } else if (ParseDate && (length == 12)) {
    /* A date! - and the length has to be 12 chars */
    OutString("\\yr");
    for (i=0;i<4;i++)
      OutChar(text[i]);
    OutString("\\mo");
    for (i=4;i<6;i++)
      OutChar(text[i]);
    OutString("\\dy");
    for (i=6;i<8;i++)
      OutChar(text[i]);
    OutString("\\hr");
    for (i=8;i<10;i++)
      OutChar(text[i]);
    OutString("\\min");
    for (i=10;i<12;i++)
      OutChar(text[i]);
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';
      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	/* Skip all unicodes that are fatcows */
	if (ucodesize > 2) {
	  i=i+ucodesize-1;
	  continue;
	}
	/*
	  For details on UTF-8 see: RFC 2044
	  Handle all ISO 8859-1 subcodes, ignore everything else
	  UTF 8 looks like this: 110xxxxx 1xxxxx
	  Where the first two 11's is the size of the ucode token
	  and xxx... the actual token. The first 1 in the second
	  byte is only there to mark it as a unicode entity.
	*/
	else {
	  i++;
	  ucode = text[i] & 0x7F;
	  ucode = ucode | ((plotchar & 0x3F) << 7);
	  ucode = ucode - 0xC0;
	  if ((ucode < 0x80) || (ucode > 0xFF))
	    continue;
	  plotchar = (unsigned char) ucode;
	  OutStringCellCheck(Escape(plotchar));
	  LastChar = plotchar;
	  continue;
	}
      }
      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ') {
	  OutCharCellCheck(plotchar);
	}
      } else {
	/* Just mash out the text ... handle backslashes */
	if (plotchar == '\\') {
	  OutStringCellCheck("\\\\");
	  LastChar = '\\';
	  continue;
	}
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' ')) {
	  OutCharCellCheck(plotchar);
	}
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}


/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  int i;
  char tempentity[STR_MAXLEN] = "";

  /* The article has to start first... */
  if (bTextThroughput) {
    if (s[0]=='&' && s[length-1]==';') {
      for (i=1;i<length-1;i++)
	tempentity[i-1]=s[i];
      tempentity[i]='\0';
      OutStringCellCheck(Translate(tempentity));
    }
    /* else just mash out the raw entity ... */
    else for (i=0;i<length;i++)
      OutCharCellCheck(s[i]);
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  /* No tool specific options */
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  /* No tool specific help */
}

/* End exported functions */
