/* xmltotex.c
   Program for parsing and converting article.dtd compliant 
   XML files to (La-)TeX format
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include <stdarg.h>
#include "common.h"

/* Language identifier "sv"=Swedish, "en"=English etc... */
char *cLangCode;
/* Parse a date */
bool ParseDate = false;
/* Has the article begun at all? */
bool bTextThroughPut = false;
/* Disable graphics by default */
bool EnableGraphics = false;
/* Flags */
bool bIsBook = false;
bool bRight = false;
bool bCenter = false;
/* Are we working with an unnumbered (bulleted) table? */
bool UnnumberedMode = false;
bool NumberedMode = false;

/*
  ROUTINES
  Building tables in memory as linked 2D matrix list
*/

typedef struct tRow tRow;
typedef struct tColumn tColumn;

struct tRow {
  tColumn *columns;
  tRow *next;
};

struct tColumn {
  char *cell;
  tColumn *next;
};

/* Global variables used for storing tables */
tRow *Table;
char *MemCell;
int bWriteCell = false;

static tRow *
newRow()
{
  tRow *newp;
  
  newp = (tRow *) malloc(sizeof(tRow));
  newp->columns = NULL;
  newp->next = NULL;
  return newp;
}

static tColumn *
newColumn(char *data)
{
  tColumn *newp;

  newp = (tColumn *) malloc(sizeof(tColumn));
  newp->cell = data;
  newp->next = NULL;
  return newp;
}

static tRow *
AddRow(tRow *table)
{
  tRow *p;

  if (table == NULL)
    return newRow();
  for (p = table; p->next != NULL; p = p->next)
    ;
  p->next = newRow();
  return table;
}

static void 
AddCell(tRow *table, char *data)
{
  tRow *p;
  tColumn *q;

  /* This first thing should actually never happen */
  if (table == NULL)
    p = newRow();
  /* But this is the normal operation */
  else
    for (p = table; p->next != NULL; p = p->next)
      ;
  if (p->columns == NULL) {
    p->columns = newColumn(data);
  } else {
    for (q = p->columns; q->next != NULL; q = q->next)
      ;
    q->next = newColumn(data);
  }
}

static void 
DeallocateTable(tRow *table)
{
  tRow *p, *r;
  tColumn *q, *s;
  
  if (table == NULL)
    return;
  p = table;
  while (p != NULL) {
    q = p->columns;
    while (q != NULL) {
      if (q->cell != NULL)
	free(q->cell);
      s = q->next;
      free(q);
      q = s;
    }
    r = p->next;
    free(p);
    p = r;
  }
}

/* Returns the max width (in columns) of a table */
static int 
GetCols(tRow *table)
{
  tRow *p;
  tColumn *q;
  register int tmp, cols;
  
  /* Could be dangerous to return 0 here... */
  if (table == NULL)
    return 1;
  p = table;
  cols = 0;
  while (p != NULL) {
    q = p->columns;
    tmp = 0;
    while (q != NULL) {
      tmp++;
      q = q->next;
    }
    if (tmp>cols)
      cols = tmp;
    p = p->next;
  }
  if (cols == 0)
    return (1);
  else
    return (cols);
}

static void 
NormalizeTable(tRow *table)
{
  tRow *p;
  tColumn *q;
  char *EmptyString;
  register int cols, tmp;

  if (table == NULL)
    return;
  cols = GetCols(table);
  p = table;

  while (p != NULL) {
    q = p->columns;
    tmp = 0;
    while (q != NULL) {
      tmp++;
      q = q->next;
    }
    q = p->columns;
    while(q->next != NULL)
      q = q->next;
    while (tmp<cols) {
      /*
	Add empty columns until we have
	equal width of all rows, some may
	need a space character in them.
      */
      EmptyString = (char *) malloc(2 * sizeof(char));
      EmptyString[0] = ' ';
      EmptyString[1] = '\0';
      q->next = newColumn(EmptyString);
      q = q->next;
      tmp ++;
    }
    p = p->next;
  }
}

static void 
PlotTable(tRow *table)
{
  tRow *p;
  tColumn *q;
  register int i, cols;

  OutString("\\medskip{}\n");
  OutString("\\begin{center}\n");
  OutString("\\begin{tabular}{");
  cols = GetCols(table);
  for (i=0; i<cols; i++)
    OutChar('l');
  OutString("}\n");
  if (table == NULL)
    return;
  p = table;
  while (p != NULL) {
    q = p->columns;
    i = 0;
    while (q != NULL) {
      if (i!=0)
	OutString(" & ");
      OutString("%s", q->cell);
      q = q->next;
      i++;
    }
    OutString("\\\\\n");
    p = p->next;
  }
  OutString("\\end{tabular}\n");
  OutString("\\end{center}\n");
  OutString("\\medskip{}\n\n\\noindent{}");  
}

static char *
OutMemS(char *AppendString, char *String)
{
  if (AppendString == NULL) {
    AppendString = malloc((strlen(String)+1) * sizeof(char));
    strcpy(AppendString, String);
  } else {
    AppendString = (char *) realloc(AppendString, (strlen(String) + strlen(AppendString) + 1) * sizeof(char) );
    strcat(AppendString, String);
  }
  return AppendString;
}

static char *
OutMemC(char *AppendString, char c)
{
  int i;
  
  if (AppendString == NULL) {
    AppendString = (char *) malloc(2 * sizeof(char));
    i = 0;
  } else {
    AppendString = (char *) realloc(AppendString, (strlen(AppendString) + 2) * sizeof(char));
    i = strlen(AppendString);
  }
  AppendString[i] = c;
  AppendString[i+1] = '\0';
  return AppendString;
}

void 
OutStringCellCheck(char *fmt, ...)
{
  char tmpBuffer[160];
  va_list ap;

  va_start(ap, fmt);
  if (!bWriteCell) {
    /* This routine just passes the string through */
    OutString(fmt, ap);
  } else {
    vsprintf(tmpBuffer, fmt, ap);
    MemCell = OutMemS(MemCell, tmpBuffer);
  }
  va_end(ap);
}

void 
OutCharCellCheck(char c)
{
  if (!bWriteCell)
    OutChar(c);
  else
    MemCell = OutMemC(MemCell, c);
}

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  int i;
  char cTmp[128];

  if (!strcmp(el,"article")) {
    bTextThroughPut = true;
    OutString("\\documentclass[a4paper]{article}\n");
    OutString("\\addtolength{\\topmargin}{-1cm}\n");
    OutString("\\addtolength{\\evensidemargin}{-3cm}\n");
    OutString("\\addtolength{\\oddsidemargin}{-1cm}\n");
    OutString("\\addtolength{\\textwidth}{4cm}\n");
  }
  else if (!strcmp(el,"book")) {
    bTextThroughPut = true;
    bIsBook = true;
    OutString("\\documentclass[a4paper]{book}\n");
    OutString("\\addtolength{\\topmargin}{-1cm}\n");
    OutString("\\addtolength{\\evensidemargin}{-3cm}\n");
    OutString("\\addtolength{\\oddsidemargin}{-1cm}\n");
    OutString("\\addtolength{\\textwidth}{4cm}\n");
  }
  else if (!strcmp(el,"metadata")) {
    if (EnableGraphics)
      OutString("\\usepackage{graphics}\n");
    for (i = 0; attr[i]; i += 2) {
      /* Handle the language somehow... */
      if (!strcmp(attr[i], "language")) {
	cLangCode = cFixLangCode(attr[i+1]);
	if (!strcmp(cLangCode,"SV")) {
	  /* Swedish language attribute */
	  OutString("\\usepackage[T1]{fontenc}\n\\usepackage[swedish]{babel}\n");
	}
      }
    }
    /* Some conditional statements for PDFTeX parsing */
    OutString("\\newif\\ifpdf\n");
    OutString("\\ifx\\pdfoutput\\undefined\n");
    OutString("\t\\pdffalse\n");
    OutString("\\else\n");
    OutString("\t\\pdftrue\n");
    OutString("\\fi\n\n");
    OutString("\\ifpdf\n");
    OutString("\t%% Saves a lot of space in PDF files\n");
    OutString("\t\\usepackage{times}\n");
    OutString("\t\\pdfcompresslevel=9\n");
    OutString("\\fi\n\n");
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = true;
    OutString("\\title{");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = true;
    OutString("\\author{");
  }
  else if (!strcmp(el,"keywords")) {
    /* OK this is one ugly fix to cut out the keywords */
    /* I don't know how to handle them right now */
    bTextThroughPut = false;
  }
  else if (!strcmp(el,"published")) {
    /* Set a flag to the text parser and handle this correctly */
    ParseDate = true;
    OutString("\\date{");
  }
  else if (!strcmp(el,"h1")) {
    PreserveWhitespace = true;
    if (!attr[0]) {
      /* Output a chapter heading */
      if (bIsBook)
	OutString("\\chapter{");
      else
	OutString("\\section*{");
    }
    /* Not implemented handling of weird justification as of now,
       but some standard ones like centering work fine */
    else for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right")) {
	  if (bIsBook)
	    OutString("\\chapter{\\raggedright ");
	  else
	    OutString("\\section*{\\raggedright ");
	} else if (!strcmp(attr[i+1], "center")) {
	  if (bIsBook)
	    OutString("\\chapter{\\centering ");
	  else
	    OutString("\\section*{\\centering ");
	} else if (!strcmp(attr[i+1], "justify")) {
	  /* Do nothing, just flush, this is the normal mode for TeX */
	  if (bIsBook)
	    OutString("\\chapter{");
	  else
	    OutString("\\section*{");
	} else {
	  /* Do nothing, just flush */
	  if (bIsBook)
	    OutString("\\chapter{");
	  else
	    OutString("\\section*{");
	}
      }
    }
  }
  /* Second level heading */
  else if (!strcmp(el,"h2")) {
    PreserveWhitespace = true;
    if (!attr[0]) {
      /* Output a chapter heading */
      if (bIsBook)
	OutString("\\section*{");
      else
	OutString("\\subsection*{");
    }
    /* Not implemented handling of weird justification as of now,
       but some standard ones like centering work fine */
    else for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right")) {
	  if (bIsBook)
	    OutString("\\section*{\\raggedright ");
	  else
	    OutString("\\subsection*{\\raggedright ");
	} else if (!strcmp(attr[i+1], "center")) {
	  if (bIsBook)
	    OutString("\\section*{\\centering ");
	  else
	    OutString("\\subsection*{\\centering ");
	} else if (!strcmp(attr[i+1], "justify")) {
	  /* Do nothing, just flush, this is the normal mode for TeX */
	  if (bIsBook)
	    OutString("\\section*{");
	  else
	    OutString("\\subsection*{");
	} else {
	  /* Do nothing, just flush */
	  if (bIsBook)
	    OutString("\\section*{");
	  else
	    OutString("\\subsection*{");
	}
      }
    }
  }
  /* Third level heading */
  else if (!strcmp(el,"h3")) {
    PreserveWhitespace = true;
    if (!attr[0]) {
      /* Output a chapter heading */
      if (bIsBook)
	OutString("\\subsection*{");
      else
	OutString("\\subsubsection*{");
    }
    /* Not implemented handling of weird justification as of now,
       but some standard ones like centering work fine */
    else for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right")) {
	  if (bIsBook)
	    OutString("\\subsection*{\\raggedright ");
	  else
	    OutString("\\subsubsection*{\\raggedright ");
	} else if (!strcmp(attr[i+1], "center")) {
	  if (bIsBook)
	    OutString("\\subsection*{\\centering ");
	  else
	    OutString("\\subsubsection*{\\centering ");
	} else if (!strcmp(attr[i+1], "justify")) {
	  /* Do nothing, just flush, this is the normal mode for TeX */
	  if (bIsBook)
	    OutString("\\subsection*{");
	  else
	    OutString("\\subsubsection*{");
	} else {
	  /* Do nothing, just flush */
	  if (bIsBook)
	    OutString("\\subsection*{");
	  else
	    OutString("\\subsubsection*{");
	}
      }
    }
  }
  /* Paragraph */
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = true;
    if (!attr[0]) {
      /* Do nothing, just flush */
    }
    /* Not implemented handling of weird justification as of now */
    else for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "align")) {
	if (!strcmp(attr[i+1], "right")) {
	  bRight = true;
	  OutString("\\begin{flushright}\n");
	} else if (!strcmp(attr[i+1], "center")) {
	  bCenter = true;
	  OutString("\\begin{center}\n");
	} else if (!strcmp(attr[i+1], "justify")) {
	  /* Do nothing, just flush, this is the normal mode for TeX */
	} else {
	  /* Do nothing, just flush */
	}
      }
    }
  }
  else if (!strcmp(el,"br"))
    OutString("\\\\\n");
  else if (!strcmp(el,"hr"))
    /* Horizontal line */
    OutString("\\medskip{}\n\n\\noindent{}");
  else if (!strcmp(el, "blockquote")) {
    PreserveWhitespace = true;
    OutString("\\begin{quotation}\n\n\\noindent{}");
  }
  /* This part handles different kinds of lists... */
  else if (!strcmp(el, "ul")) {
    UnnumberedMode = true;
    OutString("\\begin{itemize}\n\n");
  } else if (!strcmp(el, "ol")) {
    NumberedMode = true;
    OutString("\\begin{enumerate}\n\n");
  } else if (!strcmp(el, "li")) {
    PreserveWhitespace = true;
    OutString("\\item ");
  }
  /* Definition lists */
  else if (!strcmp(el, "dl")) {
    OutString("\\begin{description}\n");
  } else if (!strcmp(el, "dt")) {
    PreserveWhitespace = true;
    OutString("\\item[");
  } else if (!strcmp(el, "dd")) {
    PreserveWhitespace = true;
    OutString(" ");
  }
  /* TABLES */
  /* Make tables write to posted file, then read back */
  else if (!strcmp(el,"table")) {
    Table = NULL;
  } else if (!strcmp(el,"tr")) {
    Table = AddRow(Table);
  } else if (!strcmp(el,"td")) {
    MemCell = NULL;
    PreserveWhitespace = true;
    bWriteCell = true;
    /* Flag to send to string buffer */
  }
  /* Text formatting */
  else if (!strcmp(el,"b")) {
    OutStringCellCheck("\\textbf{");
  }
  else if (!strcmp(el,"i"))
    OutStringCellCheck("\\textit{");
  /* Images and their attributes */
  else if (!strcmp(el,"img")) {
    if (EnableGraphics) {
      OutString("\\medskip{}\n");
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "src")) {
	  /* 
	     Remap .* to .eps
	     Ultimately this section would
	     call an external program to convert
	     any image (.gif, .jpg etc) to .eps
	   */
	  OutString("\\begin{center}\n");
	  strncpy(cTmp, attr[i+1], 127);
	  cTmp[127] = '\0';
	  if (cTmp[strlen(cTmp)-4] == '.') 
	    strcpy(cTmp+strlen(cTmp)-3, "pdf");
	  OutString("\\ifpdf\n");
	  OutString("\t\\includegraphics{%s}\n", cTmp);
	  OutString("\\else\n");
	  if (cTmp[strlen(cTmp)-4] == '.')
	    strcpy(cTmp+strlen(cTmp)-3, "eps");
	  OutString("\t\\includegraphics{%s}\n", cTmp);
	  OutString("\\fi\n");
	  OutString("\\end{center}\n");
	}
	/*
	  This will only work if we put it into a {figure}
	  if (!strcmp(attr[i], "alt"))
	  OutString("\\caption{%s}\n", attr[i+1]);
	*/
      }
      OutString("\\medskip{}\n\n\\noindent{}");
    } else {
      OutString("%s: ", cGetLImage(cLangCode));
      for (i = 0; attr[i]; i += 2) {
	if (!strcmp(attr[i], "alt"))
	  OutString("%s", attr[i+1]);
      }
      OutString("\n\n");
    }
  }

  /* Footnotes */
  else if (!strcmp(el,"footnote")) {
    OutStringCellCheck("\\footnote{");
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  if (!strcmp(el,"article") ||
      !strcmp(el,"book")) {
    bTextThroughPut = false;
    OutString("\\end{document}\n");
  }
  else if (!strcmp(el,"metadata")) {
    OutString("\\begin{document}\n\\maketitle\n");
    if (bIsBook)
      OutString("\\tableofcontents\n");
    OutString("\n");
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = false;
    OutString("}\n");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = false;
    OutString("}\n");
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = false;
    /* OK this is one ugly fix */
    /* Used to cut out the keywords */
    bTextThroughPut = true;
  }
  else if (!strcmp(el,"published")) {
    ParseDate = false;
    OutString("}\n");
  }
  /* Just flowing paragraphs here */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")) {
    PreserveWhitespace = false;
    OutString("}\n\n");
  }
  else if (!strcmp(el,"p")) {
    if (bRight) {
      OutString("\\end{flushright}\n");
      bRight = false;
    }
    else if (bCenter) {
      OutString("\\end{center}\n");
      bCenter = false;
    }
    PreserveWhitespace = false;
    OutString("\n\n");
  }
  else if (!strcmp(el,"blockquote")) {
    PreserveWhitespace = false;
    OutString("\\end{quotation}\n\n\\noindent{}");
  }
  /* This section handles different kinds of lists... */
  else if (!strcmp(el,"ul")) {
    UnnumberedMode = false;
    OutString("\\end{itemize}\n\n\\noindent{}");
  } else if (!strcmp(el,"ol")) {
    NumberedMode = false;
    OutString("\\end{enumerate}\n\n\\noindent{}");
  } else if (!strcmp(el,"dl")) {
    OutString("\\end{description}\n\n");
  } else if (!strcmp(el,"li") ||
	     !strcmp(el,"dd")
	     ) {
    PreserveWhitespace = false;
    /* You know they all look the same in the end */
    OutString("\n\n");
  } else if (!strcmp(el,"dt")) {
    OutString("] ");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
    NormalizeTable(Table);
    PlotTable(Table);
    DeallocateTable(Table);
  } else if (!strcmp(el,"tr")) {
    /* Do nothing */
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = false;
    AddCell(Table, MemCell);
    bWriteCell = false;
  }
  /* Elements to be plain copied and ended WITHOUT ANY ENTER after them */
  else if (!strcmp(el,"b") ||
	   !strcmp(el,"i") ||
	   !strcmp(el,"footnote")) {
    OutStringCellCheck("}");
    PreserveWhitespace = true;
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  register unsigned int i,j;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static char temp_string_1[15];
  char *temp_string_2;
  static unsigned char LastChar = ' ';

  if (!bTextThroughPut) {
    /* Do nothing of the Article hasnt started */
  } else if (ParseDate && (length == 12)) {
    /* A date! - and the length has to be 12 chars */
    for (i=0;i<4;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=4;i<6;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=6;i<8;i++)
      OutChar(text[i]);
    OutString(" ");
    for (i=8;i<10;i++)
      OutChar(text[i]);
    OutString(":");
    for (i=10;i<12;i++)
      OutChar(text[i]);
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';

      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	for (j=0;j<ucodesize;j++) {
	  temp_string_1[j] = text[i];
	  i++;
	}
	i--;
	temp_string_1[j] = '\0';
	/* Create recoding engine and request if needed */
	if (recode_engine_outer == NULL)
	  recode_engine_outer = recode_new_outer(true);
	if (recode_engine_request_2 == NULL) {
	  recode_engine_request_2 = recode_new_request(recode_engine_outer);
	  recode_scan_request(recode_engine_request_2, "utf8..TeX");
	}
	temp_string_2 = recode_string(recode_engine_request_2, temp_string_1);
	if (temp_string_2) {
	  /* else just mash out the raw entity ... */
	  OutStringCellCheck(temp_string_2);
	  free(temp_string_2);
	  LastChar = '\0';
	  continue;
	} else {
	  plotchar = ' ';
	}
      }
      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  OutCharCellCheck(plotchar);
      } else {
	/* Just mash out the text ... handle backslashes */
	if (plotchar == '\\') {
	  OutStringCellCheck("\\backslash{}");
	  LastChar = '\\';
	  continue;
	}
	if (plotchar == '$') {
	  OutStringCellCheck("\\$");
	  LastChar = '$';
	  continue;
	}
	if (plotchar == '%') {
	  OutStringCellCheck("\\%%");
	  LastChar = '%';
	  continue;
	}
	if (plotchar == '_') {
	  OutStringCellCheck("\\_");
	  LastChar = '_';
	  continue;
	}
	if (plotchar == '#') {
	  OutStringCellCheck("\\#");
	  LastChar = '#';
	  continue;
	}
	if (plotchar == '&') {
	  OutStringCellCheck("\\&");
	  LastChar = '&';
	  continue;
	}
	if (plotchar == '{') {
	  OutStringCellCheck("\\{");
	  LastChar = '{';
	  continue;
	}
	if (plotchar == '}') {
	  OutStringCellCheck("\\}");
	  LastChar = '}';
	  continue;
	}
	if (plotchar == '<') {
	  OutStringCellCheck("$<$");
	  LastChar = '<';
	  continue;
	}
	if (plotchar == '>') {
	  OutStringCellCheck("$>$");
	  LastChar = '>';
	  continue;
	}
	/* This one surprise me a bit */
	if (plotchar == '\"') {
	  OutStringCellCheck("\"{}");
	  LastChar = '\"';
	  continue;
	}
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' '))
	  OutCharCellCheck(plotchar);
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}


/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  register int i;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;

  /* The article has to start first... */
  if (bTextThroughPut) {
    if (s[0]=='&' && s[length-1]==';') {
      /* Create recoding engine and request if needed */
      if (recode_engine_outer == NULL)
	recode_engine_outer = recode_new_outer(true);
      if (recode_engine_request_1 == NULL) {
	recode_engine_request_1 = recode_new_request(recode_engine_outer);
	recode_scan_request(recode_engine_request_1, "h4..TeX");
      }
      for(i=0;i<length;i++)
	temp_string_1[i]=s[i];
      temp_string_1[length] = '\0';
      temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
      if (temp_string_2)
	OutStringCellCheck(temp_string_2);
      free(temp_string_2);
    }
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  if (!strcmp(argv[arg],"-gphx"))
    EnableGraphics = true;
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  printf("         -gphx         = Enables graphics inclusion\n");
}

/* End exported functions */
