/* xmltotxt.c
   Program for parsing and converting article.dtd compliant 
   XML files to ISO8859-1 raw Text (TXT) format.
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 1999 Linus Walleij

This file is part of the xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "common.h"

/* Language identifier "sv"=Swedish, "en"=English etc... */
char *cLangCode;
/* Parse a date */
bool ParseDate = false;
/* Has the article begun at all? */
bool ArticleStarted = false;
/* Are we working with an unnumbered (bulleted) table? */
bool UnnumberedMode = false;
int NumberedMode = 0;
/* 
   This one counts paragraphs, used for identifying blockquotes
   or footnotes with inlined paragraphs.
*/
int ParaCounter = 0;
/* For detecting extraneous space */
static unsigned char LastChar = ' ';

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  int i;

  if (!strcmp(el,"article") ||
      !strcmp(el,"book") ) {
    ArticleStarted = true;
  }
  else if (!strcmp(el,"metadata")) {
    OutString("Article:\n");
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "cathegory"))
	OutString("        Cathegory: %s\n", attr[i+1]);
      /* Handle the language somehow... */
      if (!strcmp(attr[i], "language")) {
	cLangCode = cFixLangCode(attr[i+1]);
	OutString("        Language: %s\n", cGetLLanguage(cLangCode));
      }
    }
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = true;
    IndentChars = 21;
    OutString("        Description: ");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = true;
    IndentChars = 16;
    OutString("        Author: ");
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = true;
    IndentChars = 18;
    OutString("        Keywords: ");
  }
  else if (!strcmp(el,"published")) {
    /* Set a flag to the text parser and handle this correctly */
    ParseDate = true;
    OutString("        Published: ");
  }
  else if (!strcmp(el,"chapter")) {
    /* Add a pagebreak at chapter divisions */
    OutString("\n\n\x0C");
  }
  else if (!strcmp(el,"body")) {
    OutString("\n\n");
  }
  /* Elements to be copied */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3")
	   ) {
    PreserveWhitespace = true;
  }
  else if (!strcmp(el,"p")) {
    PreserveWhitespace = true;
    ParaCounter++;
  }
  else if (!strcmp(el,"br")) {
    OutString("\n");
    LastChar = ' ';
  }
  else if (!strcmp(el,"hr")) {
    OutString("\n\n---- * ----\n\n");
    LastChar = ' ';
  }
  else if (!strcmp(el, "blockquote")) {
    /* The blockquote is treated like a special kind of indented paragraph */
    /* Have to do something about this in the charmongler routine...       */
    IndentChars = 5;
    PreserveWhitespace = true;
    ParaCounter = 0;
    /* The first indentation manually created... */
    OutString("     ");
  }
  /* This part handles different kinds of lists... */
  else if (!strcmp(el, "ul")) {
    UnnumberedMode = true;
  } else if (!strcmp(el, "ol")) {
    /* Yeah use the flag as list enumerator... hard 2 c? */
    NumberedMode++;
  } else if (!strcmp(el, "li")) {
    PreserveWhitespace = true;
    if (UnnumberedMode)
      OutString("*  ");
    else if (NumberedMode) {
      OutString("%d. ", NumberedMode);
      NumberedMode++;
    }
    IndentChars = 3;
  }
  else if (!strcmp(el, "dl")) {
    /* Do nothing */
  } else if (!strcmp(el, "dt")) {
    PreserveWhitespace = true;
  } else if (!strcmp(el, "dd")) {
    PreserveWhitespace = true;
    OutString("        ");
    IndentChars = 8;
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
    /* Do nothing */
  } else if (!strcmp(el,"tr")) {
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = true;
  }
  /* Text formatting */
  else if (!strcmp(el,"b"))
    OutString("_");
  else if (!strcmp(el,"i"))
    OutString("*");
  /* Images and their attributes */
  else if (!strcmp(el,"img")) {
    OutString("<%s: ", cGetLImage(cLangCode));
    for (i = 0; attr[i]; i += 2) {
      if (!strcmp(attr[i], "alt"))
	OutString("%s", attr[i+1]);
    }
    OutString(">\n\n");
  }
  /* Footnotes */
  else if (!strcmp(el,"footnote")) {
    ParaCounter = 0;
    FootnoteNo++;
    OutString("(%d)", FootnoteNo);
    SetStackMode();
    /* If this is the first footnote, create a footnote stacking file */
    if (FootnoteNo == 1)
      FootnoteStackFile = tmpfile();
    OutString("(%d) ", FootnoteNo);
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  char c;

  if (!strcmp(el,"article") ||
      !strcmp(el,"book")) {
    ArticleStarted = false;
  }
  else if (!strcmp(el,"metadata")) {
    OutString("        Document created by Yelah XMLTOTXT converter\n");
    OutString("        Find us at: http://www.yelah.net/\n");
  }
  else if (!strcmp(el,"body") ||
	   !strcmp(el,"chapter")) {
    if (FootnoteNo > 0) {
      FootnoteNo = 0;
      OutString("\n\n\n%s:\n\n", cGetLFootnotes(cLangCode));
      if (fseek(FootnoteStackFile, 0, SEEK_SET)) {
	fprintf(stderr, "Could not read temporary footnote stack file.\n");
	exit(-1);
      }
      c = fgetc(FootnoteStackFile);
      while (c != EOF) {
	OutChar(c);
	c = fgetc(FootnoteStackFile);
      }
      fclose(FootnoteStackFile);
    }
    PreserveWhitespace = false;
    OutString("\n");
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = false;
    IndentChars = 0;
    OutString("\n");
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = false;
    IndentChars = 0;
    OutString("\n");
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = false;
    IndentChars = 0;
    OutString("\n");
  }
  else if (!strcmp(el,"published")) {
    ParseDate = false;
    OutString("\n");
  }
  /* Just flowing paragraphs here */
  else if (
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3") ||
	   !strcmp(el,"p")) {
    PreserveWhitespace = false;
    OutString("\n\n");
  }
  else if (!strcmp(el,"blockquote")) {
    PreserveWhitespace = false;
    IndentChars = 0;
    if (!ParaCounter)
      OutString("\n\n");
    else
      OutString("\n");
  }
  /* This section handles different kinds of lists... */
  else if (!strcmp(el,"ul")) {
    UnnumberedMode = false;
    OutString("\n");
  } else if (!strcmp(el,"ol")) {
    NumberedMode = 0;
    OutString("\n");
  } else if (!strcmp(el,"dl")) {
    OutString("\n");
  } else if (!strcmp(el,"li") ||
	     !strcmp(el,"dt") ||
	     !strcmp(el,"dd")
	     ) {
    PreserveWhitespace = false;
    IndentChars = 0;
    /* You know they all look the same in the end */
    OutString("\n");
  }
  /* TABLES */
  else if (!strcmp(el,"table")) {
  } else if (!strcmp(el,"tr")) {
    OutString("\n");
  } else if (!strcmp(el,"td")) {
    PreserveWhitespace = false;
    OutString("\t");
  }
  /* Elements to be plain copied and ended WITHOUT ANY ENTER after them */
  else if (!strcmp(el,"b"))
    OutString("_");
  else if (!strcmp(el,"i"))
    OutString("*");
  else if (!strcmp(el,"footnote")) {
    if (!ParaCounter)
      OutString("\n\n");
    UnSetStackMode();
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  register unsigned int i,j;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static char temp_string_1[15];
  char *temp_string_2;

  if (!ArticleStarted) {
    /* Do nothing of the Article hasnt started */
  } else if (ParseDate && (length == 12)) {
    /* A date! - and the length has to be 12 chars */
    for (i=0;i<4;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=4;i<6;i++)
      OutChar(text[i]);
    OutString("-");
    for (i=6;i<8;i++)
      OutChar(text[i]);
    OutString(" ");
    for (i=8;i<10;i++)
      OutChar(text[i]);
    OutString(":");
    for (i=10;i<12;i++)
      OutChar(text[i]);
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';
      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	for (j=0;j<ucodesize;j++) {
	  temp_string_1[j] = text[i];
	  i++;
	}
	i--;
	temp_string_1[j] = '\0';
	/* Create recoding engine and request if needed */
	if (recode_engine_outer == NULL)
	  recode_engine_outer = recode_new_outer(true);
	if (recode_engine_request_2 == NULL) {
	  recode_engine_request_2 = recode_new_request(recode_engine_outer);
	  recode_scan_request(recode_engine_request_2, "utf8..latin1");
	}
	temp_string_2 = recode_string(recode_engine_request_2, temp_string_1);
	if (temp_string_2) {
	  /* else just mash out the raw entity ... */
	  plotchar = temp_string_2[0];
	  free(temp_string_2);
	} else {
	  plotchar = ' ';
	}
      }
      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  OutChar(plotchar);
      } else {
	/* Just mash out the text ... */
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' ')) {
	  OutChar(plotchar);
	}
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}


/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  register int i;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;

  /* Only valid escapes */
  if (s[0]=='&' && s[length-1]==';') {
    /* Create recoding engine and request if needed */
    if (recode_engine_outer == NULL)
      recode_engine_outer = recode_new_outer(true);
    if (recode_engine_request_1 == NULL) {
      recode_engine_request_1 = recode_new_request(recode_engine_outer);
      recode_scan_request(recode_engine_request_1, "h4..latin1");
    }
    for(i=0;i<length;i++)
      temp_string_1[i]=s[i];
    temp_string_1[length] = '\0';
    temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
    if (temp_string_2) {
      /* else just mash out the raw entity ... */
      OutString(temp_string_2);
      free(temp_string_2);
    }
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  /* No tool specific options */
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  /* Implement a generic charset translator */
  /* printf("         -charset xxx  = Recode charset selection (default = iso8859-1)\n"); */
}

/* End exported functions */
