/* lcd2xml.c
   Program for parsing and converting lcdarticle.dtd or
   lcdbook.dtd compliant XML files to XML according to
   the lcdarticle.dtd or lcdbook.dtd, but with pure UTF-8
   encoding and no excessive linefeeds etc.
   Copyright (C) 1999 Linus Walleij

This file is part of the LCDtools package.

LCDtools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "common.h"

static unsigned char LastChar = ' ';
static unsigned idepth = 0;
/* For handling of headerlessness */
bool Header = true;
/* DTD URI to use on XML documents */
char DTDURI[128] = "\0";
/* A stylesheet if the user wants one */
char CSSURI[128] = "\0";

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  register int i;

  if (!strcmp(el,"article")) {
    LineBreakMode = false;
    idepth = 0;
    if (Header) {
      if (!strlen(DTDURI))
	strcpy(DTDURI, "lcdarticle.dtd");
      OutString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
      OutString("<!DOCTYPE article SYSTEM \"%s\">\n", DTDURI);
      if (strlen(CSSURI))
	OutString("<?xml-stylesheet href=\"%s\" type=\"text/css\"?>\n", CSSURI);
      OutString("<article>\n");
      CharThroughPut = true;
    }
  }
  else if (!strcmp(el,"book")) {
    LineBreakMode = false;
    idepth = 0;
    if (Header) {
      if (!strlen(DTDURI))
	strcpy(DTDURI, "lcdbook.dtd");
      OutString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
      OutString("<!DOCTYPE book SYSTEM \"%s\">\n", DTDURI);
      if (strlen(CSSURI))
	OutString("<?xml-stylesheet href=\"%s\" type=\"text/css\"?>\n", CSSURI);
      OutString("<book>\n");
      CharThroughPut = true;
    }
  }

  /* Empty */
  else if (!strcmp(el,"br") ||
	   !strcmp(el,"hr") ||
	   !strcmp(el,"img")) {
    OutString("<%s", el);
    for (i = 0; attr[i]; i += 2) {
      OutString(" %s=\"%s\"", attr[i], attr[i+1]);
    }
    OutString(" />\n");
    LastChar = ' ';
  }
  
  /* To be followed by ENTER */
  else if (!strcmp(el,"metadata") ||
	   !strcmp(el,"blockquote") ||
	   !strcmp(el,"table") ||
	   !strcmp(el,"body") ||
	   !strcmp(el,"chapter") ||
	   !strcmp(el,"tr") ||
	   !strcmp(el,"ol") ||
	   !strcmp(el,"ul") ||
	   !strcmp(el,"dl")) {
    idepth ++;
    OutString("<%s", el);
    PreserveWhitespace = true;
    for (i = 0; attr[i]; i += 2) {
      OutString(" %s=\"%s\"", attr[i], attr[i+1]);
    }
    OutString(">\n");
    LastChar = ' ';
  }

  /* Not to be followed by ENTER */
  else if (!strcmp(el,"description") ||
	   !strcmp(el,"author") ||
	   !strcmp(el,"keywords") ||
	   !strcmp(el,"published") ||
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3") ||
	   !strcmp(el,"p") ||
	   !strcmp(el,"b") ||
	   !strcmp(el,"i") ||
	   !strcmp(el,"li") ||
	   !strcmp(el,"dt") ||
	   !strcmp(el,"dd") ||
	   !strcmp(el,"td") ||
	   !strcmp(el,"a") ||
	   !strcmp(el,"footnote") ) {
    idepth ++;
    OutString("<%s", el);
    PreserveWhitespace = true;
    for (i = 0; attr[i]; i += 2) {
      OutString(" %s=\"%s\"", attr[i], attr[i+1]);
    }
    OutString(">");
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  if (!strcmp(el,"article")) {
    OutString("</article>\n");
    CharThroughPut = false;
  }
  else if (!strcmp(el,"book")) {
    OutString("</book>\n");
    CharThroughPut = false;
  }
  
  /* To be followed by ENTER */
  else if (!strcmp(el,"metadata") ||
	   !strcmp(el,"blockquote") ||
	   !strcmp(el,"table") ||
	   !strcmp(el,"tr") ||
	   !strcmp(el,"ol") ||
	   !strcmp(el,"ul") ||
	   !strcmp(el,"dl") ||
	   !strcmp(el,"description") ||
	   !strcmp(el,"author") ||
	   !strcmp(el,"keywords") ||
	   !strcmp(el,"published") ||
	   !strcmp(el,"body") ||
	   !strcmp(el,"chapter") ||
	   !strcmp(el,"h1") ||
	   !strcmp(el,"h2") ||
	   !strcmp(el,"h3") ||
	   !strcmp(el,"p") ||
	   !strcmp(el,"li") ||
	   !strcmp(el,"dt") ||
	   !strcmp(el,"dd") ||
	   !strcmp(el,"td")) {
    idepth ++;
    PreserveWhitespace = false;
    OutString("</%s>\n", el);
  }

  /* Not to be followed by ENTER */
  else if (
	   !strcmp(el,"b") ||
	   !strcmp(el,"i") ||
	   !strcmp(el,"a") ||
	   !strcmp(el,"footnote") ) {
    idepth ++;
    PreserveWhitespace = true;
    OutString("</%s>", el);
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  register unsigned int i,j;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static char temp_string_1[15];

  if (!CharThroughPut) {
    /* Do nothing of the throughput hasnt started */
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';
      
      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	for (j=0;j<ucodesize;j++) {
	  temp_string_1[j] = text[i];
	  i++;
	}
	i--;
	temp_string_1[j] = '\0';
	if (CharThroughPut)
	  OutString(temp_string_1);
	LastChar = '\0';
	continue;
      }

      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  OutChar(plotchar);
      } else {
	/* Just mash out the text ... */
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' ')) {
	  /* Some HTML characters require special treatment */
	  if (plotchar == '<')
	    OutString("&lt;");
	  else if (plotchar == '>')
	    OutString("&gt;");
	  else if (plotchar == '&')
	    OutString("&amp;");
	  else
	    OutChar(plotchar);
	}
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}

/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  register int i;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;
  
  /* The article has to start first... */
  if (CharThroughPut) {
    if ((s[0]=='&' && s[length-1]==';') &&
	strcmp(s, "&quot;") &&
	strcmp(s, "&lt;") &&
	strcmp(s, "&gt;") ) {
      /* Create recoding engine and request if needed */
      if (recode_engine_outer == NULL)
	recode_engine_outer = recode_new_outer(true);
      if (recode_engine_request_1 == NULL) {
	recode_engine_request_1 = recode_new_request(recode_engine_outer);
	recode_scan_request(recode_engine_request_1, "h4..utf8");
      }
      for(i=0;i<length;i++)
	temp_string_1[i]=s[i];
      temp_string_1[length] = '\0';
      temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
      if (temp_string_2)
	OutString(temp_string_2);
      free(temp_string_2);
    } else {
      /* Just mash out the raw entity ... */
      for (i=0;i<length;i++)
	OutChar(s[i]);
      LastChar = '\0';
    }
  }
}

/* scans for tool-specific options and sets variables correspondingly */
void
ScanToolOptions(const char **argv, int arg) {
  if (!strcmp(argv[arg],"-noheader"))
    Header = false;
  if (!strcmp(argv[arg],"-dtd")) {
    if(*argv[arg+1] == '-') {
      fprintf(stderr, "You must give a <dtduri> parameter to the -dtd switch!\n");
      exit(-1);
    }
    strncpy(DTDURI, argv[arg+1], 127);
    DTDURI[127] = '\0';
    fprintf(stderr, "Using DTD URI: \"%s\"\n", DTDURI);
  }
  if (!strcmp(argv[arg],"-css")) {
    if(*argv[arg+1] == '-') {
      fprintf(stderr, "You must give a <cssuri> parameter to the -css switch!\n");
      exit(-1);
    }
    strncpy(CSSURI, argv[arg+1], 127);
    CSSURI[127] = '\0';
    fprintf(stderr, "Using CSS URI: \"%s\"\n", CSSURI);
  }
}

/* Print any tool specific help */
void
PrintToolHelp(void) {
  printf("         -noheader     = Write no headers\n");
  printf("         -dtd <dtduri> = Use this DTD URI string on XML documents\n");
  printf("                         long URI:s may be \"quoted\"\n");
  printf("         -css <cssuri> = Use this CSS URI string on XML documents\n");
  printf("                         long URI:s may be \"quoted\"\n");
}

/* End exported functions */
