/* xmlindexer.c
   Program for indexing all xmlfiles
   Requires Expat (commonly packaged as xmltok-dev) and MySQL interfaces
   Copyright (C) 1999 Linus Walleij

This file is part of the Yelah xmltools package.

Yelah xmltools is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "config.h"

#if HAVE_STDBOOL_H
# include <stdbool.h>
#else
typedef enum {false = 0, true = 1} bool;
#endif
#include <stdio.h>
#if HAVE_STDLIB_H
# include <stdlib.h>
#endif
#include <dirent.h>
#include <sys/stat.h>
#if HAVE_EXPAT_H
# include <expat.h>
#else
# include <xmlparse.h>
#endif
#include <recode.h>
#if HAVE_STRING_H
# include <string.h>
#endif
#include <mysql.h>
#define STR_MAXLEN      255
#define BUFFSIZE        8192

/* Global variables */
char Buff[BUFFSIZE];

/* MySQL connection object */
MYSQL mysql;

/* MySQL connection strings and such */
char MySQLhost[50] = "localhost";
char MySQLdatabase[50] = "YELAH_DATABASE";
char MySQLuser[50] = "yelahadmin";
char MySQLpassword[50] = "";

/* SQL stuff to actually print to the table at end */
/* Double field lengths in some cases because of possible escapes */
char ArticleID[50] = "";
char CathegoryLabel[50] = ""; /* Used to look up CATHEGORYID */
char CathegoryID[10] = "";    /* This one */
char Published[20] = "";      /* A date */
char Language[3] = "";        /* Language code 2 chars */
char Keywords[512] = "";      /* Keywords */
char Description[512] = "";   /* Description of this article */
char Author[512] = "";        /* The author of this article */

/* Recode engines and requests */
RECODE_OUTER recode_engine_outer = NULL;
RECODE_REQUEST recode_engine_request_1 = NULL;
/* Identifier for recode, must have program name */
const char program_name[]= PACKAGE "-" VERSION;

/* Used for temporary encoding */
char EscapeBuffer[512] = "";

/* This is like a "file" the parser writes to */
char * OutString;

/* Parse a date */
bool ParseDate = false;
/* Preserve Whitespace in the current area? */
bool PreserveWhitespace = false;
/* Has the metadata begun at all? */
bool Metadata = false;
/* After metadata is read, skip the rest of the file */
bool MetadataBuffered = false;


void
StringPutc(char ToPut) {
  int OutLength;

  OutLength = strlen(OutString);
  OutString[OutLength] = ToPut;
  OutString[OutLength+1] = '\0';
}

/* Start handler for XML elements */
void 
startElementHandler(void *data, const XML_Char *el, const XML_Char **attr)
{
  int i;

  if (!strcmp(el,"metadata")) {
    Metadata = true;
    for (i = 0; attr[i]; i += 2) {
      /* Handle the cathegory */
      if (!strcmp(attr[i], "cathegory"))
	strncpy(CathegoryLabel,attr[i+1],50);
      /* Handle the language */
      if (!strcmp(attr[i], "language")) {
	strncpy(Language,attr[i+1],2);
      }
    }
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = true;
    OutString = Description;
    OutString[0] = '\0';
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = true;
    OutString = Author;
    OutString[0] = '\0';
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = true;
    OutString = Keywords;
    OutString[0] = '\0';
  }
  else if (!strcmp(el,"published")) {
    /* Set a flag to the text parser and handle this correctly */
    ParseDate = true;
    OutString = Published;
    OutString[0] = '\0';
  }
}


/* End handler for XML elements */
void 
endElementHandler(void *data, const XML_Char *el)
{
  if (!strcmp(el,"metadata")) {
    Metadata = false;
    MetadataBuffered = true;
  }
  else if (!strcmp(el,"description")) {
    PreserveWhitespace = false;
  }
  else if (!strcmp(el,"author")) {
    PreserveWhitespace = false;
  }
  else if (!strcmp(el,"keywords")) {
    PreserveWhitespace = false;
  }
  else if (!strcmp(el,"published")) {
    ParseDate = false;
  }
}


/* The raw text falls through to here */
void 
textHandler(void *data, const XML_Char *text, int length)
{
  unsigned int i;
  unsigned int ucode;
  unsigned int ucodesize;
  unsigned char plotchar;
  unsigned char tempChar;
  static unsigned char LastChar = ' ';

  if (!Metadata) {
    /* Do nothing if the Metadata isnt active */
  } else if (ParseDate && (length == 12)) {
    /* A date! - and the length has to be 12 chars */
    for (i=0;i<4;i++)
      StringPutc(text[i]);
    StringPutc('-');
    for (i=4;i<6;i++)
      StringPutc(text[i]);
    StringPutc('-');
    for (i=6;i<8;i++)
      StringPutc(text[i]);
  } else {
    for (i=0;i<length;i++) {
      plotchar = (unsigned char) text[i];
      /* Less than space? Ingore it. */
      if (plotchar < (unsigned char) ' ')
	plotchar = ' ';
      /* Unicode UTF-8 handling */
      if (plotchar > 0x7F) {
	tempChar = plotchar;
        ucodesize = 0;
	while (tempChar & 0x80) {
	  tempChar = tempChar << 1;
	  ucodesize++;
	}
	/* Skip all unicodes that are fatcows */
	if (ucodesize > 2) {
	  i=i+ucodesize-1;
	  continue;
	}
	/*
	  For details on UTF-8 see: RFC 2044
	  Handle all ISO 8859-1 subcodes, ignore everything else
	  UTF 8 looks like this: 110xxxxx 1xxxxx
	  Where the first two 11's is the size of the ucode token
	  and xxx... the actual token. The first 1 in the second
	  byte is only there to mark it as a unicode entity.
	*/
	else {
	  i++;
	  ucode = text[i] & 0x7F;
	  ucode = ucode | ((plotchar & 0x3F) << 7);
	  ucode = ucode - 0xC0;
	  if ((ucode < 0x80) || (ucode > 0xFF))
	    continue;
	  plotchar = (unsigned char) ucode;
	}
      }
      /* Then put it on screen */
      if (!PreserveWhitespace) {
	/* Remove whitespace and everything below x20 hex */
	if (plotchar > (unsigned char) ' ')
	  StringPutc(plotchar);
      } else {
	/* Just mash out the text ... */
	/* Remove extraneous space, tabs, linebreaks */
	if (!(LastChar == ' ' && plotchar == ' '))
	  StringPutc(plotchar);
	LastChar = plotchar;
      }
    } /* End for each character loop */
  }
}


/* All raw entities are handled here */
void 
defaultHandler(void *userData, const XML_Char *s, int length)
{
  register int i;
  static char temp_string_1[15]; /* How long can an escape code be? */
  char *temp_string_2;
  
  /* The article has to start first... */
  if (Metadata) {
    if (s[0]=='&' && s[length-1]==';') {
      /* Create recoding engine and request if needed */
      if (recode_engine_outer == NULL)
	recode_engine_outer = recode_new_outer(true);
      if (recode_engine_request_1 == NULL) {
	recode_engine_request_1 = recode_new_request(recode_engine_outer);
	recode_scan_request(recode_engine_request_1, "h4..latin1");
      }
      for(i=0;i<length;i++)
	temp_string_1[i]=s[i];
      temp_string_1[length] = '\0';
      temp_string_2 = recode_string(recode_engine_request_1, temp_string_1);
      if (temp_string_2) {
	/* else just mash out the raw entity ... */
	strcat(OutString, temp_string_2);
	free(temp_string_2);
      }
    }
  }
}

/************************************************************************
 * Parse a file referenced by a file pointer
 * Return true on successful parse, false on failed.
 ************************************************************************/
bool 
Parse(FILE * ParseFile)
{
  XML_Parser parser;
  int done;
  int len;
  /* Signifies if the program should write to stdout instead of a file */

  parser = XML_ParserCreate(NULL);

  if (! parser) {
    fprintf(stdout, "?Out of memory error?\n");
    return(false);
  }

  XML_SetElementHandler(parser, startElementHandler, endElementHandler);
  XML_SetCharacterDataHandler(parser, textHandler);
  XML_SetDefaultHandler(parser, defaultHandler);

  done = false;
  MetadataBuffered = false;
  while (!done && !MetadataBuffered) {
    len = fread(Buff, 1, BUFFSIZE, ParseFile);
    if (ferror(ParseFile)) {
      fprintf(stdout, "Could not read from input file!\n");
      return(false);
    }
    done = feof(ParseFile);

    if (! XML_Parse(parser, Buff, len, done)) {
      fprintf(stdout, "Parse error at line %d:\n%s\n",
	      XML_GetCurrentLineNumber(parser),
	      XML_ErrorString(XML_GetErrorCode(parser)));
      XML_ParserFree(parser);
      return (false);
    }
  }
  XML_ParserFree(parser);
  return(true);
}

/************************************************************************
 * Recursively search files and directories below the directory i
 * "base_path" for the file "name". Return a open file pointer to the
 * file in question or NULL if the search failed.
 ************************************************************************/
void 
FileScan(char * base_path)
{
  DIR * DirPointer;
  struct dirent * Entry;
  struct stat EntryStatus;
  char cTempPath[512];
  char cTempFname[512];
  char query[8192];
  int bIsDir;
  int bIsFile;
  int nTemp;
  FILE * TempPointer=NULL;
  MYSQL_RES * res;
  MYSQL_ROW row;

  DirPointer = opendir(base_path);
  if (! DirPointer){
    fprintf(stdout, "Unable to open directory %s.\n", base_path);
    return;
  }
  while ((Entry = readdir(DirPointer)) != NULL)
    {
      strncpy(cTempPath, base_path, 512);
      /* Sometimes the trailing slash is there, sometimes it's not */
      if (cTempPath[strlen(cTempPath)-1] != '/')
	strcat(cTempPath, "/");
      strcat(cTempPath, Entry->d_name);
      if (stat(cTempPath, &EntryStatus)) {
	fprintf(stdout, "Failed to read file status for %s.\n", cTempPath);
	return;
      }
      bIsDir = S_ISDIR(EntryStatus.st_mode);
      bIsFile = S_ISREG(EntryStatus.st_mode);
      if (bIsDir && strcmp(Entry->d_name, ".") && strcmp(Entry->d_name,"..") )
	FileScan(cTempPath);
      if (bIsFile) {
	strcpy(cTempFname, Entry->d_name);
	if (!strcmp(cTempFname+strlen(cTempFname)-4, ".xml")) {
	  /* Stuff to do when found */
	  /* TempPointer to be used */
	  strncpy(ArticleID, cTempFname, 50);
	  ArticleID[strlen(ArticleID)-4] = '\0';
	  /* Uncomment this to list all processed files */
	  fprintf(stdout, "Parsing: %s\n", cTempPath);
	  TempPointer = fopen(cTempPath, "r");
	  if (TempPointer != NULL) {
	    if (!Parse(TempPointer)) {
	      fprintf(stdout, "Failed to parse: %s\n", cTempPath);
	    }

	    /* If it was successfully parsed */
	    else {
	      /* SQL ENCODE ALL STRINGS -- they may contain apostrophes etc */
	      mysql_escape_string(EscapeBuffer, ArticleID, strlen(ArticleID));
	      strncpy(ArticleID, EscapeBuffer, 50);
	      mysql_escape_string(EscapeBuffer, Keywords, strlen(Keywords));
	      strncpy(Keywords, EscapeBuffer, 512);
	      mysql_escape_string(EscapeBuffer, Author, strlen(Author));
	      strncpy(Author, EscapeBuffer, 512);
	      mysql_escape_string(EscapeBuffer, Description, strlen(Description));
	      strncpy(Description, EscapeBuffer, 512);
	      
	      /* GET THE CATHEGORY ID FROM THE LABEL */
	      sprintf(query, "SELECT CATHEGORYID FROM CATHEGORIES WHERE LABEL=\"%s\"", CathegoryLabel);
	      if(mysql_query(&mysql,query)) {
		fprintf(stdout,"QUERY: %s\nSELECTION: Query failed (%s)\n",query,mysql_error(&mysql));
		exit(1);
	      }
	      if (!(res=mysql_use_result(&mysql))) {
		fprintf(stdout,"Couldn't get CathegoryID from database: %s\n",
			mysql_error(&mysql));
		exit(1);
	      }
	      /* Get the info from this field ... */
	      row = (MYSQL_ROW) mysql_fetch_row(res);
	      
	      /*******************************************************
	       * If the cathegory exists, then insert in archive. 
	       * otherwise the program will create and add the new
	       * cathegory. The following rows create a new ID
	       *******************************************************/
	      
	      if (row == NULL) {
		
		mysql_free_result(res);
		sprintf(query, "SELECT CATHEGORYID FROM CATHEGORIES ORDER BY CATHEGORYID DESC");
		if(mysql_query(&mysql,query)) {
		  fprintf(stdout,"QUERY: %s\nSELECTION: Query failed (%s)\n",query,mysql_error(&mysql));
		  exit(1);
		}
		if (!(res=mysql_use_result(&mysql))) {
		  fprintf(stdout,"Couldn't get CathegoryID from database: %s\n",
			  mysql_error(&mysql));
		  exit(1);
		}
		/* Get the index and increase by one */
		row = (MYSQL_ROW) mysql_fetch_row(res);
		nTemp = atoi((char *) row[0]);
		/* Loop through records or MySQL 3.21 will squeal */
		while (mysql_fetch_row(res) != NULL) {}
		mysql_free_result(res);
		nTemp++;
		sprintf(query, "INSERT INTO CATHEGORIES (CATHEGORYID, LABEL, DESCRIPTION) VALUES (%d, \"%s\", \"%s\")", nTemp, CathegoryLabel, CathegoryLabel);
		if(mysql_query(&mysql,query)) {
		  fprintf(stdout,"QUERY: %s\nINSERTION: Query failed (%s)\n",query,mysql_error(&mysql));
		  exit(1);
		}
		
		/* Then this just repeats the procedure from above */
		sprintf(query, "SELECT CATHEGORYID FROM CATHEGORIES WHERE LABEL=\"%s\"", CathegoryLabel);
		if(mysql_query(&mysql,query)) {
		  fprintf(stdout,"QUERY: %s\nSELECTION: Query failed (%s)\n",query,mysql_error(&mysql));
		  exit(1);
		}
		if (!(res=mysql_use_result(&mysql))) {
		  fprintf(stdout,"Couldn't get CathegoryID from database: %s\n",
			  mysql_error(&mysql));
		  exit(1);
		}
		row = (MYSQL_ROW) mysql_fetch_row(res);
		
		/*******************************************************
		 * If the cathegory exists, then insert in archive. 
		 *******************************************************/
		
	      } else {
		/* The cathegory should be the first field in the row */
		strcpy(CathegoryID, (char *) row[0]);
		/* This function must be called until it return NULL when using a row or... */
		while (mysql_fetch_row(res) != NULL) {}
		/* Always free the result */
		mysql_free_result(res);
		
		/* INSERT THE ARTICLE INTO THE ARCHIVE_ARTICLES TABLE */
		sprintf(query, "INSERT INTO ARCHIVE_ARTICLES VALUES (\"%s\", %s, \"%s\", \"%s\", \"%s\", \"%s\", \"%s\")", ArticleID, CathegoryID, Published, Language, Keywords, Description, Author);
		/* printf("Executing SQL Query: %s\n", query); */
		if(mysql_query(&mysql,query)) {
		  fprintf(stdout,"QUERY: %s\nINSERTION: Query failed (%s)\n",query,mysql_error(&mysql));
		}
		fclose(TempPointer);
	      }
	    }
	  } else {
	    fprintf(stdout, "Could not open file %s.\n", cTempFname);
	  }
	}
      }
    }
  closedir(DirPointer);
}

void printHelp(char *cmdline){
  printf("%s\n", program_name);
  printf("Usage: %s [-options] <directory>\n\n", cmdline);
  printf("-?, --help      this help text\n");
  printf("-h <host>       database host\n");
  printf("-d <database>   database to connect to\n");
  printf("-u <user>       login as this user\n");
  printf("-p <password>   use this password\n");
}

/************************************************************************
 * Main function call to XML translators
 ************************************************************************/
int 
main(int argc, char **argv) {
  int arg;

  if (argc<2) {
    printHelp(argv[0]);
    exit(0);
  }
  
  for (arg=1; arg<argc; arg++) {
    if (*argv[arg] == '-') {
      /* -? or -help: show the command line help page */
      if (!strcmp(argv[arg],"-?") || 
          !strcmp(argv[arg],"-help") || 
          !strcmp(argv[arg],"--help")) {
	printHelp(argv[0]);
        exit(0);
      }
      /* MySQL host parameter */
      if (!strcmp(argv[arg],"-h")) {
        if(*argv[arg+1] == '-') {
          fprintf(stderr, "You must give a <hostname> parameter to the -h switch!\n");
          exit(-1);
        }
        strcpy(MySQLhost, argv[arg+1]);
        fprintf(stderr, "Using MySQL host: %s\n", MySQLhost);
      }
      /* MySQL user parameter */
      if (!strcmp(argv[arg],"-u")) {
        if(*argv[arg+1] == '-') {
          fprintf(stderr, "You must give a <username> parameter to the -u switch!\n");
          exit(-1);
        }
        strcpy(MySQLuser, argv[arg+1]);
        fprintf(stderr, "Logging in as MySQL user: %s\n", MySQLuser);
      }
      /* MySQL password parameter */
      if (!strcmp(argv[arg],"-p")) {
        if(*argv[arg+1] == '-') {
          fprintf(stderr, "You must give a <password> parameter to the -p switch!\n");
          exit(-1);
        }
        strcpy(MySQLpassword, argv[arg+1]);
        fprintf(stderr, "Using MySQL password: ********\n");
      }
      /* MySQL database parameter */
      if (!strcmp(argv[arg],"-d")) {
        if(*argv[arg+1] == '-') {
          fprintf(stderr, "You must give a <database> parameter to the -d switch!\n");
          exit(-1);
        }
        strcpy(MySQLdatabase, argv[arg+1]);
        fprintf(stderr, "Connecting to MySQL database: %s\n", MySQLdatabase);
      }
    }
  }

  /* CONNECT TO THE DATABASE */
#if MYSQL_VERSION_ID > 32133
  mysql_init(&mysql);
  if (!(mysql_real_connect(&mysql, MySQLhost, MySQLuser, MySQLpassword, MySQLdatabase, 0, 0, 0))) {
#else
  if (!(mysql_real_connect(&mysql, MySQLhost, MySQLuser, MySQLpassword, 0, NULL, 0))) {
#endif
    fprintf(stdout,"%s\n",mysql_error(&mysql));
    fprintf(stdout,"Host:     %s\n", MySQLhost);
    fprintf(stdout,"Database: %s\n", MySQLdatabase);
    fprintf(stdout,"User:     %s\n", MySQLuser);
    fprintf(stdout,"Password: %s\n", MySQLpassword);
    exit(1);
  }
  /* USE THE YELAH_DATABASE */
#if MYSQL_VERSION_ID <= 32133
  /* mysql_select_db(&mysql, MySQLdatabase); */
#endif

  /* Previous method
  if(mysql_query(&mysql,"USE YELAH_DATABASE")) {
    fprintf(stdout,"USING: Query failed (%s)\n",mysql_error(&mysql));
  }
  */

  /* DELETE ALL ARTICLES FROM THE ARCHIVE_ARTICLES TABLE */
  if(mysql_query(&mysql,"DELETE FROM ARCHIVE_ARTICLES")) {
    fprintf(stdout,"DELETION: Query failed (%s)\n",mysql_error(&mysql));
  }

  FileScan(argv[argc-1]);
  
  mysql_close(&mysql);

  /* Free recode engine */
  if (recode_engine_request_1 != NULL)
    recode_delete_request(recode_engine_request_1);
  if (recode_engine_outer != NULL)
    recode_delete_outer(recode_engine_outer);
  
  exit(0);
}
